"""Asynchronous USB driver for the Crazyradio USB dongle."""

from aiocflib.utils.usb import (
    claim_device,
    find_devices,
    get_vendor_setup,
    is_pyusb1,
    release_device,
    send_vendor_setup,
    USBDevice,
    USBError,
)
from anyio import Lock, to_thread
from array import array
from binascii import hexlify
from contextlib import asynccontextmanager, AsyncExitStack
from enum import IntEnum
from functools import total_ordering, wraps
from sys import exc_info
from typing import Iterable, List, Optional, Tuple, Union

from aiocflib.utils.addressing import (
    CrazyradioAddress,
    CrazyradioAddressLike,
    CrazyradioDataRate,
    DEFAULT_RADIO_ADDRESS as DEFAULT_ADDRESS,
    parse_radio_uri,
    to_radio_address,
)
from aiocflib.utils.concurrency import Full, ThreadContext
from aiocflib.errors import NotFoundError

__author__ = "CollMot Robotics Ltd"
__all__ = ("Crazyradio", "DEFAULT_ADDRESS")


CRADIO_VID = 0x1915
CRADIO_PID = 0x7777


class CrazyradioConfigurationRequest(IntEnum):
    """Enum representing the configuration requests supported by the radio.

    See http://wiki.bitcraze.se/projects:crazyradio:protocol for documentation
    """

    SET_RADIO_CHANNEL = 0x01
    SET_RADIO_ADDRESS = 0x02
    SET_DATA_RATE = 0x03
    SET_RADIO_POWER = 0x04
    SET_RADIO_ARD = 0x05
    SET_RADIO_ARC = 0x06
    ACK_ENABLE = 0x10
    SET_CONT_CARRIER = 0x20
    SCAN_CHANNELS = 0x21
    LAUNCH_BOOTLOADER = 0xFF


class CrazyradioPower(IntEnum):
    """Enum representing the power levels supported by the radio."""

    P_M18DBM = 0
    P_M12DBM = 1
    P_M6DBM = 2
    P_0DBM = 3


def _find_devices(serial: Optional[str] = None) -> List[USBDevice]:
    """Returns a list of Crazyradio dongles currently connected to the computer.

    This function uses `pyusb` functions directly, hence it will block the
    calling thread. You must call it in a separate worker thread when you use
    it in conjunction with an asyncio framework.

    Parameters:
        serial: when specified, returns only those devices that match the given
            serial number. (Typically there will be only a single device).

    Returns:
        the list of Crazyradio dongles
    """
    candidates = find_devices(vid=CRADIO_VID, pid=CRADIO_PID)
    if serial is not None:
        candidates = [dev for dev in candidates if dev.serial_number == serial]
    return candidates


def get_serials() -> Tuple[str]:
    """Returns the serial numbers of all the connected Crazyradio dongles.

    Returns:
        the serial numbers of all the connected Crazyradio dongles
    """
    return tuple(dev.serial_number for dev in _find_devices())


class Acknowledgment:
    """Simple value object representing an acknowledgment from the radio."""

    @classmethod
    def from_array(cls, data: array, arc: int):
        """Constructs an acknowledgment from the raw bytes received from the
        radio via the USB connection.

        Parameters:
            data: the byte array that was received
            arc: the value of the ACK retry count in the connection; used when
                the input indicates that the packet was not acknowledged
        """
        result = cls()
        if data[0] != 0:
            result.ack = bool(data[0] & 0x01)
            result.power_detector_status = bool(data[0] & 0x02)
            result.retry_count = data[0] >> 4
            result.data = bytes(data[1:])
        else:
            result.retry_count = arc
        return result

    def __init__(
        self,
        data: bytes = b"",
        *,
        ack: bool = False,
        power_detector_status: bool = False,
        retry_count: int = 0,
    ):
        self.ack = ack
        self.power_detector_status = power_detector_status
        self.retry_count = retry_count
        self.data = data

    def __repr__(self) -> str:
        return (
            "{0.__class__.__name__}({0.data!r}, ack={0.ack!r}, "
            "power_detector_status={0.power_detector_status!r}, "
            "retry_count={0.retry_count!r})"
        ).format(self)


@total_ordering
class RadioConfiguration:
    """Simple value class that contains the commonly used configuration variables
    of a Crazyradio instance that we expect the user to provide when sending
    packets.

    This class is also used to specify a single configuration of a radio scan
    for reachable devices.
    """

    @classmethod
    def ensure(cls, value: "RadioConfigurationLike"):
        """When the input is a string, converts it into a RadioConfiguration_
        object, assuming that it contains a ``radio://`` URI. When the input
        is a RadioConfiguration_, it returns the object intact.

        Raises:
            TypeError: if the input is neither a string nor a RadioConfiguration
        """
        if isinstance(value, str):
            return cls.from_uri(value)
        elif isinstance(value, cls):
            return value
        else:
            raise TypeError("expected string or {0}, got {1!r}", cls.__name__, value)

    @classmethod
    def from_uri(cls, uri: str):
        """Creates a RadioConfiguration_ object from a ``radio://`` URI.

        The scheme of the supplied URI is ignored; it is assumed that the rest
        of the URI follows the format used for ``radio://`` URIs, and that the
        first path component is the radio index (which will also be ignored).
        """
        parts = parse_radio_uri(uri)
        del parts["index"]
        return cls(**parts)

    def __init__(
        self,
        *,
        address: Optional[CrazyradioAddressLike] = None,
        channel: Optional[int] = None,
        data_rate: CrazyradioDataRate = CrazyradioDataRate.DR_2MPS,
        scheme: str = "radio",
    ):
        """Constructor.

        Parameters:
            address: the address to use for sending packets
            data_rate: the data rate to use for sending packets
            channel: the channel to use for sending packets; `None` is usable
                for radio channel scans where we want to specify that the
                scan will happen on all channels. `None` is not allowed if the
                object is used for declaring where packets should be sent.
            scheme: the URI scheme to use when converting the configuration to
                a URI
        """
        self._address = Crazyradio.to_address(address) if address is not None else None
        self._channel = channel
        self._data_rate = data_rate
        self._scheme = str(scheme)

    @property
    def address(self) -> Optional[CrazyradioAddress]:
        """The address to use when sending packets."""
        return self._address

    @property
    def channel(self) -> Optional[int]:
        """The channel to use when sending packets."""
        return self._channel

    @property
    def data_rate(self) -> CrazyradioDataRate:
        """The data rate to use when sending packets."""
        return self._data_rate

    @property
    def is_full(self) -> bool:
        """Returns whether the configuration is fully specified."""
        return (
            self._address is not None
            and self._channel is not None
            and self._data_rate is not None
            and self._scheme is not None
        )

    @property
    def scheme(self) -> str:
        """The URI scheme to use when converting this configuration to a URI."""
        return self._scheme

    def replace(
        self,
        address: Optional[CrazyradioAddressLike] = None,
        data_rate: Optional[CrazyradioDataRate] = None,
        channel: Optional[int] = None,
        scheme: Optional[str] = None,
    ):
        """Replaces the address, the data rate and/or the channel in the
        configuration object and returns a new configuration object.
        """
        return self.__class__(
            address=address if address is not None else self._address,
            data_rate=data_rate if data_rate is not None else self._data_rate,
            channel=channel if channel is not None else self._channel,
            scheme=scheme if scheme is not None else self._scheme,
        )

    def to_uri(self, index: int = 0) -> str:
        """Converts the radio configuration to a string URI that can be
        passed to a CRTPDevice_ constructor.
        """
        parts = ["{0}://{1}".format(self.scheme, index)]
        if self.channel is not None:
            parts.append(str(self.channel))
            if self.data_rate is not None:
                parts.append(str(self.data_rate))
                if self.address is not None:
                    parts.append(hexlify(self.address).upper().decode("ascii"))
        return "/".join(parts)

    def __eq__(self, other):
        return (self._data_rate, self._channel, self._address, self._scheme) == (
            other._data_rate,
            other._channel,
            other._address,
            other._scheme,
        )

    def __lt__(self, other):
        # Order is important here: we want RadioConfiguration objects to be
        # sortable in a way that same data rates are clustered together. This
        # is because it is faster to switch addresses than channels or data
        # rates, and we want the radio scans to be as fast as possible.
        return (self._data_rate, self._channel, self._address, self._scheme) < (
            other._data_rate,
            other._channel,
            other._address,
            other._scheme,
        )

    def __repr__(self):
        return (
            "{0.__class__.__name__}(address={0.address!r}, "
            "channel={0.channel!r}, "
            "data_rate={0.data_rate!r}, "
            "scheme={0.scheme!r})"
        ).format(self)


#: Type specification for objects that can be converted into a RadioConfiguration
RadioConfigurationLike = Union[RadioConfiguration, str]


class Crazyradio:
    """Low-level driver object that is used for communication with the
    Crazyflie via a Crazyradio.

    This object is intended to be used as an asynchronous context manager as
    follows::

        device = await Crazyradio.detect_one()
        async with device as radio:
            response = await radio.send_and_receive_bytes(b"1234")
            if response and response.ack:
                print(repr(response.data))

    In the context, a background thread is running and managing the low-level
    communication with the device; you may call `radio.send_and_receive_bytes()`
    to send raw bytes to and receive raw bytes from the device. The thread is
    terminated when the execution exits the context.

    Note that you cannot receive data from the Crazyflie without sending some
    as the downstream is contained in the acknowledgment packets. If you need
    to receive data but you have nothing to say, send a CRTP null packet.

    If there is another thread that is already using the same device, the
    context blocks upon entering until the device becomes available.
    """

    @classmethod
    async def detect_all(cls):
        """Creates a list of low-level driver objects by scanning the USB buses
        for a suitable USB dongle.
        """
        devices = await to_thread.run_sync(_find_devices)
        return [cls(device) for device in devices]

    @classmethod
    async def detect_one(cls, *, index: int = 0):
        """Creates a low-level driver object by scanning the USB buses for a
        suitable USB dongle, and selecting one based on the provided device
        index.

        Parameters:
            index: the index of the USB device to select if multiple dongles
                are connected

        Raises:
            NotFoundError: if there is no such device with the given index
        """
        devices = await to_thread.run_sync(_find_devices)
        try:
            device = devices[index]
        except IndexError:
            raise NotFoundError() from None
        return cls(device)

    @classmethod
    async def from_uri(cls, uri: str):
        """Creates a low-level driver object from its URI representation.

        Only the device index is used from the URI; the remaining parts are
        ignored.
        """
        parts = parse_radio_uri(uri)
        return await cls.detect_one(index=parts["index"])

    to_address = to_radio_address

    def __init__(self, device: USBDevice):
        """Constructor.

        Creates a low-level driver object that uses the specified USB device.

        Parameters:
            device: the USB device to use
        """
        self._device = device

        self._arc = -1  # type: int
        self._current_address = None  # type: Optional[CrazyradioAddress]
        self._current_channel = None  # type: Optional[int]
        self._current_configuration = None  # type: Optional[RadioConfiguration]
        self._current_data_rate = None  # type: Optional[CrazyradioDataRate]
        self._version = None  # type: Optional[float]

        self._sender_thread_context = ThreadContext.create_worker(
            setup=self._configure_device, teardown=self._teardown_device
        )

        self._exit_stack = None  # type: Optional[AsyncExitStack]

    @property
    def version(self) -> Optional[float]:
        """Returns the version number of the associated device.

        This property returns a valid value only after you have connected to
        the USB device.
        """
        return self._version

    async def __aenter__(self):
        """Opens the driver object. This function must be called before you
        start using the driver.

        Starts an OS-level thread that will be responsible for managing
        communication over the given low-level device. Returns when the
        thread has been started.
        """
        assert self._exit_stack is None

        exit_stack = AsyncExitStack()

        try:
            await exit_stack.__aenter__()
            await exit_stack.enter_async_context(claim_device(self._device))
            sender = await exit_stack.enter_async_context(self._sender_thread_context)
            self._exit_stack = exit_stack
        finally:
            if self._exit_stack is None:
                await exit_stack.__aexit__(*exc_info())

        return _CfRadioCommunicator(sender, self)

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        assert self._exit_stack is not None
        exit_stack = self._exit_stack
        self._exit_stack = None
        return await exit_stack.__aexit__(exc_type, exc_value, tb)

    def _configure_device(self):
        """Configures the USB device when the worker thread starts.

        This function is executed in the worker thread.
        """
        device = self._device

        self._current_configuration = None
        self._version = None

        if is_pyusb1:
            try:
                cfg = device.get_active_configuration()
            except USBError:
                cfg = None
            if cfg is None or cfg.bConfigurationValue != 1:
                device.set_configuration(1)
            handle = device
            version = float(
                "{0:x}.{1:x}".format(device.bcdDevice >> 8, device.bcdDevice & 0x0FF)
            )
        else:
            handle = device.open()
            handle.setConfiguration(1)
            handle.claimInterface(0)
            version = float(device.deviceVersion)

        if version < 0.3:
            raise RuntimeError("This driver requires Crazyradio firmware V0.3+")

        self._handle, self._version = handle, version

        # Reset the dongle to power up settings
        self._set_data_rate(CrazyradioDataRate.DR_2MPS)
        self._set_channel(2)
        self._arc = -1
        if version >= 0.4:
            self._set_cont_carrier(False)
            self._set_address(b"\xe7" * 5)
            self._set_power(CrazyradioPower.P_0DBM)
            self._set_arc(3)
            self._set_ard_bytes(32)
            self._set_ack_enable(True)

    def _teardown_device(self, exc_type, exc_value, tb):
        """Tears down the connection to the USB device when the worker thread
        exits.

        This function is executed in the worker thread.
        """
        self._current_address = None
        self._current_channel = None
        self._current_configuration = None
        self._current_data_rate = None
        self._version = None

        try:
            if self._use_crtp_to_usb:
                self._set_crtp_to_usb(False)
        except Exception:
            # maybe the device was disconnected already?
            pass

        try:
            release_device(self._handle)
        finally:
            self._handle = None
            self._version = None

    def _configure(self, configuration: RadioConfiguration) -> None:
        """Sets the address, channel and data rate of the radio in a single
        call.

        This function also caches the last configuration object it was called
        with. If the current configuration object is the same as the last one
        and there were no manual changes to the address, channel and data
        rates in the meanwhile with the appropriate methods, the configuration
        step will be skipped to save some USB bandwidth.
        """
        if configuration is not self._current_configuration:
            self._set_data_rate(configuration.data_rate)
            self._set_channel(configuration.channel)
            self._set_address(configuration.address)
            self._current_configuration = configuration

    def _has_fw_scan(self):
        """Returns whether the Crazyradio supports accelerated firmware-driven
        channel scans.
        """
        return self._version is not None and self._version >= 0.5

    def _set_ack_enable(self, enable: bool) -> None:
        """Sets whether acknowledgments are enabled on the radio."""
        send_vendor_setup(
            self._handle, CrazyradioConfigurationRequest.ACK_ENABLE, int(bool(enable))
        )

    def _set_address(self, address: CrazyradioAddress) -> None:
        """Sets the address that the radio uses when sending packets.

        Parameters:
            address: the radio address to set
        """
        if len(address) != 5:
            raise ValueError("the radio address must be 5 bytes long")

        if address != self._current_address:
            send_vendor_setup(
                self._handle,
                CrazyradioConfigurationRequest.SET_RADIO_ADDRESS,
                0,
                0,
                address,
            )
            self._current_address = address
            self._current_configuration = None

    def _set_arc(self, arc: int) -> None:
        """Sets the ACK retry count in a synchronous manner.

        Parameters:
            arc: the ACK retry count to use for subsequent packets
        """
        send_vendor_setup(
            self._handle, CrazyradioConfigurationRequest.SET_RADIO_ARC, arc
        )
        self._arc = arc

    def _set_ard_bytes(self, nbytes):
        """Sets the ACK retry delay time for radio communication, in
        terms of ACK payload bytes. The firmware will calculate and use the
        equivalent number of microseconds depending on the current data rate.

        Parameters:
            us: the ACK retry delay time to use for subsequent packets, in
                terms of ACK payload bytes
        """
        send_vendor_setup(
            self._handle, CrazyradioConfigurationRequest.SET_RADIO_ARD, 0x80 | nbytes
        )

    def _set_ard_time(self, us: int) -> None:
        """Sets the ACK retry delay time for radio communication, in
        microseconds.

        Parameters:
            us: the ACK retry delay time to use for subsequent packets, in
                microseconds
        """
        # Auto Retransmit Delay:
        # 0000 - Wait 250uS
        # 0001 - Wait 500uS
        # 0010 - Wait 750uS
        # ........
        # 1111 - Wait 4000uS

        # Round down, to value representing a multiple of 250uS
        t = max(min(int((us / 250) - 1), 0xF), 0)
        send_vendor_setup(self._handle, CrazyradioConfigurationRequest.SET_RADIO_ARD, t)

    def _set_channel(self, channel: int) -> None:
        """Sets the radio channel to be used in a synchronous manner.

        Parameters:
            channel: the channel on which subsequent packets will be sent
        """
        if channel < 0 or channel > 125:
            raise ValueError("Invalid channel: {0}".format(channel))
        if channel != self._current_channel:
            send_vendor_setup(
                self._handle, CrazyradioConfigurationRequest.SET_RADIO_CHANNEL, channel
            )
            self._current_channel = channel
            self._current_configuration = None

    def _set_cont_carrier(self, active):
        """Enables or disables continuous carrier mode on the radio.

        In continuous carrier mode the radio transmit a constant sine wave at
        the currently set frequency (channel) and power. This is a test mode
        that can affect other 2.4GHz devices. It should only be used in a lab
        for testing purposes.
        """
        send_vendor_setup(
            self._handle,
            CrazyradioConfigurationRequest.SET_CONT_CARRIER,
            int(bool(active)),
        )

    def _set_data_rate(self, data_rate: CrazyradioDataRate) -> None:
        """Sets the radio data rate to be used in a synchronous manner.

        Parameters:
            data_rate: the data rate to use for subsequent packets
        """
        if data_rate != self._current_data_rate:
            send_vendor_setup(
                self._handle, CrazyradioConfigurationRequest.SET_DATA_RATE, data_rate
            )
            self._current_data_rate = data_rate
            self._current_configuration = None

    def _set_power(self, power: CrazyradioPower) -> None:
        """Sets the radio power to be used in a synchronous manner.

        Parameters:
            power: the transmission power to use for subsequent packets
        """
        send_vendor_setup(
            self._handle, CrazyradioConfigurationRequest.SET_RADIO_POWER, power
        )

    def _scan(
        self,
        targets: Optional[List[RadioConfigurationLike]] = None,
        address: Optional[
            Union[CrazyradioAddressLike, Iterable[CrazyradioAddressLike]]
        ] = None,
        packet: bytes = b"\xff\xff\xff",
    ) -> List[RadioConfiguration]:
        """Scans a selected combination of channels and data rates to detect
        devices listening on these channels.

        Parameters:
            targets: items specifying the data rates and channels to scan, or
                strings that contain ``radio://`` URIs that can be converted
                into data rates and channels. When omitted, it defaults to all
                combinations of channels and data rates.
            address: Crazyradio address to use when sending packets. `None`
                means to use the current address. A single address means to use
                the given address. An iterable of addresses means to scan all
                specified addresses. Integers are mapped to the five byte
                long address E7E7E7E7XX where XX is replaced by the integer.
            packet: packet to send during testing; defaults to the null CRTP
                packet, repeated three times

        Returns:
            a list containing the targets where a device was detected
        """
        result = []

        # If no targets are given, scan all channels and all data rates
        if targets is None:
            targets = [
                RadioConfiguration(data_rate=data_rate)
                for data_rate in CrazyradioDataRate
            ]
        else:
            targets = [RadioConfiguration.ensure(target) for target in targets]

        # If an address is given, replace all address-less targets with the
        # given address. If multiple addresses are given, replace all address-less
        # targets with a combination of each such target and address
        if address is not None:
            addresses = (
                list(address)
                if not isinstance(address, (CrazyradioAddress, int))
                else [address]
            )
            addresses = [Crazyradio.to_address(address) for address in addresses]

            new_targets = []
            for target in targets:
                if target.address is None:
                    new_targets.extend(
                        target.replace(address=address) for address in addresses
                    )
                else:
                    new_targets.append(target)

            targets = new_targets
            address = None

        # Check whether the target list contains a mixture of address-less and
        # address-based targets as this is not allowed
        has_addresses = any(target.address is not None for target in targets)
        if has_addresses:
            if any(target.address is None for target in targets):
                raise ValueError(
                    "mixing address-less targets with address-based targets is not supported"
                )

        for target in sorted(targets):
            if has_addresses:
                self._set_address(target.address)

            self._set_data_rate(target.data_rate)

            if target.channel is None:
                matches = self._scan_channels(packet=packet)
                result.extend(target.replace(channel=channel) for channel in matches)
            else:
                self._set_channel(target.channel)
                status = self._send_and_receive_bytes(packet)
                if status and status.ack:
                    result.append(target)

        return result

    def _scan_channels(
        self, first: int = 0, last: int = 125, packet: bytes = b"\xff\xff\xff"
    ) -> List[int]:
        """Scans all channels in the given channel range to detect devices
        listening on these channels.

        If the radio supports it, the scan will be executed by the firmware,
        switching channels as necessary. If the radio does not support it, the
        scan will fall back to a slower USB-driven scan where we explicitly
        switch channels one by one.

        The data rate of the radio is left untouched during the scan.

        Parameters:
            first: first channel to test (inclusive)
            last: last channel to test (inclusive)
            packet: packet to send during testing; defaults to the null CRTP
                packet, repeated three times

        Returns:
            a list containing the indices of all the channels where a device
            was detected
        """

        if self._has_fw_scan():  # Fast firmware-driven scan
            self.current_channel = None
            self.current_address = None
            self.current_data_rate = None

            send_vendor_setup(
                self._handle,
                CrazyradioConfigurationRequest.SCAN_CHANNELS,
                first,
                last,
                packet,
                timeout=2000,  # apprently 1000 msec is not enough sometimes
            )
            result = list(
                get_vendor_setup(
                    self._handle, CrazyradioConfigurationRequest.SCAN_CHANNELS, 0, 0, 64
                )
            )

            # Workaround for USB bug, see Crazyradio issue #9
            if len(result) == 64:
                result.clear()
            return result

        else:  # Slow PC-driven scan
            result = []
            for i in range(first, last + 1):
                self._set_channel(i)
                status = self._send_and_receive_bytes(packet)
                if status and status.ack:
                    result.append(i)
            return result

    def _send_and_receive_bytes(self, data: bytes) -> Optional[Acknowledgment]:
        """Sends some data via the radio connection in a synchronous manner.

        This function is executed in the worker thread.

        Returns:
            the acknowledgment received from the radio, or `None` if no
            acknowledgment was received in time

        Raises:
            IOError: when the Crazyflie was disconnected
        """
        try:
            if is_pyusb1:
                self._handle.write(endpoint=1, data=data, timeout=1000)
                response = self._handle.read(0x81, 64, timeout=1000)
            else:
                self._handle.bulkWrite(1, data, 1000)
                response = self._handle.bulkRead(0x81, 64, 1000)
            return Acknowledgment.from_array(response, arc=self._arc)
        except USBError:
            return None

    def _configure_send_and_receive_bytes(
        self, configuration: RadioConfiguration, data: bytes
    ) -> Optional[Acknowledgment]:
        """Shortcut for a common use-case that arises frequently: configure the
        radio link, then send a single packet and wait for the acknowledgment.

        This function is executed in the worker thread.

        Returns:
            the acknowledgment received from the radio, or `None` if no
            acknowledgment was received in time

        Raises:
            IOError: when the Crazyflie was disconnected
        """
        self._configure(configuration)
        return self._send_and_receive_bytes(data)


class _CfRadioCommunicator:
    """Object that is returned when entering a Crazyradio context and that allows
    us to send packets to and receive packets from the radio connection.

    The object essentially proxies all relevant methods to the underlying
    Crazyradio_ object such that the methods are executed by the worker thread.

    This is an internal class; you do not need to construct it yourself.
    """

    def __init__(self, sender, radio: Crazyradio):
        """Constructor.

        Parameters:
            sender: an async function that can be used to send a method execution
                request to the outbound worker thread
            radio: the Crazyradio object that constructed this instance
        """

        self._configuration_lock = Lock()
        self._radio = radio
        self._sender = sender

        def create_proxy_for(name):
            target = getattr(radio, "_" + name)

            @wraps(target)
            async def proxy(*args, **kwds):
                try:
                    return await sender(target, *args, **kwds)
                except Full:
                    raise IOError(
                        "Request queue to radio outbound thread is full"
                    ) from None

            return proxy

        methods = (
            "configure_send_and_receive_bytes",
            "scan",
            "scan_channels",
            "send_and_receive_bytes",
            "set_ack_enable",
            "set_arc",
            "set_ard_bytes",
            "set_ard_time",
            "set_cont_carrier",
            "set_power",
        )

        for name in methods:
            setattr(self, name, create_proxy_for(name))

    @asynccontextmanager
    async def configure(self, configuration: RadioConfiguration):
        """Configures the radio address, channel and data rate according to
        the given configuration object, and establishes a context. While the
        context is open, any other tasks trying to call `configure()` will
        block upon entering the context.

        This ensures that tasks do not step on each other's foot by modifying
        the address, channel or data rate without other tasks knowing about it.

        Parameters:
            configuration: the configuration to establish
        """
        async with self._configuration_lock:
            try:
                await self._sender(self._radio._configure, configuration)
            except Full:
                raise IOError(
                    "Request queue to radio outbound thread is full"
                ) from None
            yield


async def test():
    device = await Crazyradio.detect_one()
    async with device as radio:
        targets = await radio.scan(address=1)
        if not targets:
            print("No Crazyflie found")
        else:
            # \xfd\x01 sends a "get version" command to the link control port
            async with radio.configure(targets[0]):
                response = await radio.send_and_receive_bytes(b"\xfd\x01")
                print(repr(response))


if __name__ == "__main__":
    import trio

    trio.run(test)



from __future__ import annotations

from anyio import (
    create_memory_object_stream,
    move_on_after,
    sleep,
    Event,
    WouldBlock,
)
from contextlib import asynccontextmanager
from dataclasses import dataclass, field, replace
from functools import partial
from operator import attrgetter
from sys import exc_info
from typing import Callable, Dict, Optional, Tuple, List

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.drivers.crazyradio import (
    Acknowledgment,
    Crazyradio,
    CrazyradioAddress,
    RadioConfiguration,
    _CfRadioCommunicator,
)
from aiocflib.utils.addressing import parse_radio_uri
from aiocflib.utils.concurrency import create_daemon_task_group, gather, ObservableValue
from aiocflib.utils.statistics import SlidingWindowMean

from ..exceptions import WrongURIType
from ..strategies import (
    BackoffPollingStrategy,
    DefaultPollingStrategy,
    DefaultResendingStrategy,
    NoPollingStrategy,
    PatientResendingStrategy,
    PollingStrategy,
    ResendingStrategy,
)

from .base import CRTPDriver
from .registry import register

__all__ = ("RadioDriver",)


_instances: Dict[int, "_SharedCrazyradioState"] = {}


@dataclass
class _SharedCrazyradioState:
    radio: Optional[Crazyradio] = None
    instance: Optional[_CfRadioCommunicator] = None
    count: int = 0

    _initializing_event: Event = field(default_factory=Event)
    _destroying_event: Optional[Event] = None

    @property
    def destroying(self) -> bool:
        return self._destroying_event is not None

    async def initialize(self, index: int) -> None:
        assert self.radio is None and self.instance is None
        assert self.count == 0

        radio = await Crazyradio.detect_one(index=index)
        instance = await radio.__aenter__()

        self.radio = radio
        self.instance = instance

        self._initializing_event.set()

    def invalidate(self) -> None:
        self.invalidated = True

    def start_destruction(self) -> Tuple[Crazyradio, Event]:
        assert self.radio is not None
        assert self.count == 1
        assert self._destroying_event is None

        radio = self.radio
        self.radio = None
        self.instance = None
        self.count = 0
        self._destroying_event = Event()

        return radio, self._destroying_event

    async def wait_until_initialized(self) -> None:
        await self._initializing_event.wait()

    async def wait_until_destroyed(self) -> None:
        if self._destroying_event:
            await self._destroying_event.wait()


@asynccontextmanager
async def SharedCrazyradio(index: int):
    global _instances

    while True:
        state = _instances.get(index)
        if state is None:
            # This radio was not used yet so we need to initialize it
            _instances[index] = _SharedCrazyradioState()
            await _instances[index].initialize(index)
        elif state.destroying:
            # This radio is being destroyed in another task so let's wait until
            # it is released and then try to acquire it again
            await state.wait_until_destroyed()
        else:
            # This radio is either already initialized or is currently being
            # initialized so we are okay
            break

    # Wait until the radio is initialized
    await state.wait_until_initialized()
    state.count += 1

    try:
        yield state.instance
    finally:
        # When using Trio, we are not allowed to start any additional cancel
        # scopes here, otherwise Trio will complain about a "corrupted cancel
        # scope stack".
        state = _instances[index]
        if state.count > 1:
            state.count -= 1
        elif state.count == 1:
            radio, event = state.start_destruction()
            try:
                await radio.__aexit__(*exc_info())
            finally:
                _instances.pop(index)
            event.set()
        else:
            raise RuntimeError("Corrupted radio reference counter")


#: Type specification for radio driver presets
RadioDriverPreset = Tuple[
    Callable[[PollingStrategy], None], Callable[[ResendingStrategy], None]
]


@dataclass
class _EnabledAcquired:
    enabled: bool
    acquired: bool


class _SafeLinkState:
    """Private class that stores the current state of the safe link mode."""

    def __init__(self):
        """Constructor."""
        self._up, self._down = 8, 4
        self._enabled_acquired = ObservableValue(_EnabledAcquired(False, False))

    @property
    def acquired(self) -> bool:
        """Returns whether the safe link mode has been acquired."""
        return self._enabled_acquired.value.acquired

    @property
    def counter_bits(self) -> Tuple[int, int]:
        """Returns the current state of the counter bits, for debugging purposes."""
        return self._up, self._down

    async def disable(self) -> None:
        """Disables the safe link mode on the Crazyflie. Note that the Crazyflie
        will still operate in safe link mode if it has already been acquired,
        but it will not be a requirement any more that the safe link mode must
        be enabled at all times.
        """
        if self.enabled:
            value = self._enabled_acquired.value
            await self._enabled_acquired.set(replace(value, enabled=False))

    @property
    def enabled(self) -> bool:
        """Returns whether the safe link mode _should_ be enabled on the
        Crazyflie. This does not necessarily mean that it _is_ enabled; it only
        means that we are trying to negotiate the safe link mode with the
        Crazyflie.
        """
        return self._enabled_acquired.value.enabled

    async def notify_acquired(self) -> None:
        """Notifies the state object that the safe link mode has been
        successfully negotiated with the Crazyflie."""
        await self._set_acquired(True)

    async def notify_lost(self) -> None:
        """Notifies the state object that the safe link mode has been lost."""
        await self._set_acquired(False)

    async def enable(self) -> None:
        """Enables the safe link mode on the Crazyflie."""
        if not self.enabled:
            value = self._enabled_acquired.value
            await self._enabled_acquired.set(replace(value, enabled=True))

    def encode(self, packet: CRTPPacket) -> bytes:
        """Encodes the given CRTP packet, incorporating the safe link status
        bits in the header.
        """
        return packet.to_bytes(safelink_bits=self._up + self._down)

    def observe(self) -> ObservableValue[_EnabledAcquired]:
        """Returns an observable value that reports whether the safe link mode
        is currently enabled and acquired or not.
        """
        return self._enabled_acquired

    async def _set_acquired(self, value: bool) -> None:
        """Sets the 'acquired' flag of the safe link mode."""
        value = bool(value)

        if self.acquired == value:
            return

        if value:
            self._up, self._down = 0, 0
        else:
            self._up, self._down = 8, 4

        state = self._enabled_acquired.value
        await self._enabled_acquired.set(replace(state, acquired=value))

    def update(self, response: Acknowledgment) -> None:
        """Processes an acknowledgment received from the peer."""
        if not self.acquired:
            return

        if not response.ack:
            return

        self._up = 8 - self._up
        if response.data and response.data[0] & 0x04 == self._down:
            self._down = 4 - self._down

    async def wait_until_acquired(self) -> None:
        """Waits until the safe-link mode is acquired."""
        await self._enabled_acquired.wait_until(attrgetter("acquired"))


#: Dictionary mapping human-readable names to pairs or polling and resending
#: strategies that can be used by a radio driver
RADIO_DRIVER_PRESETS = {
    "default": (DefaultPollingStrategy, DefaultResendingStrategy),
    "patient": (BackoffPollingStrategy, PatientResendingStrategy),
    "noPolling": (NoPollingStrategy, DefaultResendingStrategy),
}


@register("radio")
class RadioDriver(CRTPDriver):
    """CRTP driver that allows us to communicate with a Crazyflie via a
    Crazyradio dongle.

    Attributes:
        polling_strategy: a callable that decides how often we should poll the
            downlink if there are no packets that we want to send to the
            Crazyflie
        resending_strategy: a callable that decides whether we should resend the
            last packet if it failed or whether we should drop the connection
    """

    PRESETS = RADIO_DRIVER_PRESETS

    @asynccontextmanager
    async def _connected_to(self, uri: str):
        try:
            parts = parse_radio_uri(uri)
        except ValueError:
            raise WrongURIType from None

        index = parts.pop("index")

        self._configuration = RadioConfiguration(**parts)

        async with SharedCrazyradio(index) as radio:
            async with create_daemon_task_group() as task_group:
                task_group.start_soon(self._worker, radio)
                task_group.start_soon(self._safe_link_supervisor, radio)
                yield self

    def __init__(self, preset: str = "default"):
        """Constructor.

        Parameters:
            preset: name of a preset from the PRESETS attribute of the class
                that determines how often the driver should poll the downlink
                with null packets and how it should handle packet resending
        """
        self._configuration = None
        self._link_quality = ObservableValue(0.0)
        self._safe_link_state = _SafeLinkState()

        try:
            self.apply_preset(preset)
        except KeyError:
            raise ValueError(f"No such preset: {preset}") from None

        # TODO(ntamas): what if the in_queue is full?
        self._in_queue_tx, self._in_queue_rx = create_memory_object_stream[CRTPPacket](
            256
        )
        self._out_queue_tx, self._out_queue_rx = create_memory_object_stream[
            CRTPPacket
        ](1)

    @property
    def address(self) -> Optional[CrazyradioAddress]:
        """The address that the driver will be configured for, or ``None`` if
        the driver has no URI.
        """
        if not self._uri:
            return None

        try:
            config = parse_radio_uri(self._uri)
        except Exception:
            return None

        return config["address"] if config else None

    def apply_preset(self, name: str) -> None:
        """Applies a preset strategy to the given connection to control how
        often should the driver pull the downlink with null packets and how it
        should handle acknowledgment failures.

        This method can be called with an active connection; the new preset
        will take effect as soon as possible.
        """
        try:
            preset = self.PRESETS[name]
        except KeyError:
            raise KeyError("no such preset: {0}".format(name)) from None

        self.polling_strategy, self.resending_strategy = preset[0](), preset[1]()

    @property
    def configuration(self) -> Optional[RadioConfiguration]:
        """The address, channel and data rate that the driver is configured for,
        or ``None`` if the driver is not configured.
        """
        return self._configuration

    @property
    def is_safe(self) -> bool:
        return self._safe_link_state.acquired

    @property
    def link_quality(self) -> ObservableValue[float]:
        return self._link_quality

    @property
    def name(self) -> str:
        return "radio"

    async def notify_rebooted(self) -> None:
        safe_link_was_enabled = self._safe_link_state.enabled
        if not safe_link_was_enabled:
            # Nothing to do
            return
        else:
            # Disable the safe link mode temporarily until the Crazyflie boots
            await self._safe_link_state.disable()
            await self._notify_safe_link_lost()
            # Wait for the Crazyflie to boot
            await sleep(1)
            # Notify the driver that it is now safe to re-enable the safe link mode
            await self._safe_link_state.enable()

    async def receive_packet(self) -> CRTPPacket:
        """Receives a single CRTP packet.

        Returns:
            the next CRTP packet that was received
        """
        return await self._in_queue_rx.receive()

    async def send_packet(self, packet: CRTPPacket) -> None:
        """Sends a CRTP packet.

        Parameters:
            packet: the packet to send
        """
        await self._out_queue_tx.send(packet)

    @classmethod
    async def scan_interfaces(
        cls, address: Optional[CrazyradioAddress] = None
    ) -> List[str]:
        """Scans all interfaces of this type for available Crazyflie quadcopters
        and returns a list with appropriate connection URIs that could be used
        to connect to them.

        Parameters:
            address: the address of the Crazyflie to look for; `None` means to
                use the default address

        Returns:
            the list of connection URIs where a Crazyflie was detected; an empty
            list is returned for interfaces that do not support scanning
        """
        devices = await Crazyradio.detect_all()
        results: List[str] = await gather(
            (cls._scan_single_interface, device, address) for device in devices
        )
        return sum(results, [])

    async def use_safe_link(self) -> None:
        """Instructs the driver to start using safe-link mode to ensure
        guaranteed packet delivery to the remote peer.
        """
        await self._safe_link_state.enable()

    async def _notify_safe_link_lost(self) -> None:
        """Notifies the driver that the established safe link state has been
        lost and it should re-establish the safe link state as soon as possible.
        """
        await self._safe_link_state.notify_lost()

    @classmethod
    async def _scan_single_interface(
        cls, radio: Crazyradio, address: Optional[CrazyradioAddress] = None
    ) -> List[str]:
        """Scans a single interface for available Crazyflie quadcopters and
        returns a list with appropriate connection URIs that could be used
        to connect to them.

        Parameters:
            radio: the radio device to use for scanning
            address: the address of the Crazyflie to look for; `None` means to
                use the default address

        Returns:
            the list of connection URIs where a Crazyflie was detected; an empty
            list is returned for interfaces that do not support scanning
        """
        async with radio as device:
            return await device.scan(address=address)

    async def _safe_link_supervisor(self, radio: Crazyradio) -> None:
        """Worker task that ensures that the radio is in safe link mode when it
        should be in safe link mode.
        """
        async for state in self._safe_link_state.observe():
            if state.enabled and not state.acquired:
                success = False
                while not success:
                    async with radio.configure(self._configuration):
                        success = await self._try_to_acquire_safe_link_mode(radio)
                    if not success:
                        await sleep(0.25)

    async def _try_to_acquire_safe_link_mode(self, radio: Crazyradio):
        """Attempts to acquire safe link mode on the Crazyflie found at the
        address, channel and data rate that the radio is currently configured to.

        Returns:
            whether the safe link mode was acquired successfully
        """
        safe_link_packet = CRTPPacket.safe_link().to_bytes()

        for _ in range(10):
            response = await radio.send_and_receive_bytes(safe_link_packet)
            if response and response.data == safe_link_packet:
                await self._safe_link_state.notify_acquired()
                return True

        return False

    async def _worker(self, radio: Crazyradio) -> None:
        """Worker task that runs continuously and handles the sending and
        receiving of packets between a given Crazyradio instance and a single
        Crazyflie (or other CRTP device).

        Parameters:
            radio: the Crazyradio instance to use
        """
        if self._safe_link_state.enabled:
            # Wait at most 2 seconds for safe-link mode before proceeding
            # without it
            with move_on_after(2):
                await self._safe_link_state.wait_until_acquired()

        null_packet = outbound_packet = CRTPPacket.null()
        delay_before_next_null_packet = 0.01

        link_quality_estimator = SlidingWindowMean(100)

        while True:
            to_send = self._safe_link_state.encode(outbound_packet)
            response = await radio.configure_send_and_receive_bytes(
                self._configuration, to_send
            )

            if response is None:
                # Resend immediately
                continue

            # Update the safe-link state
            self._safe_link_state.update(response)

            # Link quality is determined as the mean of the score of the
            # last 100 packets, where the score is determined as follows.
            # The score of a packet is 10 if it went through the first
            # time we tried to send it (retry count is zero). The score
            # decreases by 1 for every retry, and also by 1 if the
            # packet was not acknowledged at the end (meaning that it
            # got lost). Then the score is divided by 10 so we get a
            # mean link quality between 0 and 1.
            score = 9 - response.retry_count + int(response.ack)
            link_quality_estimator.add(score)
            await self._link_quality.update(link_quality_estimator.mean / 10.0)

            # Check whether the packet has to be re-sent
            action = self.resending_strategy(response.ack, to_send)
            if action == "accept":
                # Accept the packet, no resending needed
                pass
            elif action == "stop":
                # Bail out -- too many packets lost
                raise IOError("Too many packets lost")
            elif action == 0:
                # Resend immediately. If the packet that we are trying to send
                # is a filler (null) packet, poll the outbound packet queue and
                # check whether we can send something useful instead of just
                # pulling the downstream
                if outbound_packet is null_packet:
                    try:
                        outbound_packet = self._out_queue_rx.receive_nowait()
                    except WouldBlock:
                        outbound_packet = null_packet
                continue
            elif action > 0:
                # Wait a bit before resending. If the packet that we are trying
                # to send is a filler (null) packet, poll the outbound packet
                # queue after the delay and check whether we can send something
                # useful instead of just pulling the downstream
                if outbound_packet is null_packet:
                    with move_on_after(action):
                        outbound_packet = await self._out_queue_rx.receive()
                else:
                    await sleep(action)
                continue
            else:
                # Invalid response, resend immediately
                continue

            # No resending needed, process response and get next packet to send
            if response.data:
                inbound_packet = CRTPPacket.from_bytes(response.data)
                await self._in_queue_tx.send(inbound_packet)

            # Figure out how much to wait before the next null packet is sent
            delay_before_next_null_packet = self.polling_strategy(
                response.data, to_send
            )
            if delay_before_next_null_packet > 0:
                # Wait for a given number of seconds
                outbound_packet = null_packet
                with move_on_after(delay_before_next_null_packet):
                    outbound_packet = await self._out_queue_rx.receive()
            elif delay_before_next_null_packet < 0:
                # Wait indefinitely
                outbound_packet = await self._out_queue_rx.receive()
            else:
                # Poll the outbound queue; send a null packet if the queue is
                # empty
                try:
                    outbound_packet = self._out_queue_rx.receive_nowait()
                except WouldBlock:
                    outbound_packet = null_packet


register("bradio")(partial(RadioDriver, preset="patient"))



"""Superclass for devices that use the CRTP protocol for communication."""

from anyio import move_on_after
from contextlib import AsyncExitStack
from sys import exc_info
from typing import AsyncIterable, Iterable, Optional, Union

from .crtpstack import (
    CRTPDispatcher,
    CRTPPacket,
    CRTPCommandLike,
    CRTPDataLike,
    CRTPPortLike,
)
from .drivers import CRTPDriver

from aiocflib.errors import TimeoutError
from aiocflib.utils.concurrency import (
    create_daemon_task_group,
    DaemonTaskGroup,
    TaskStartedNotifier,
)

__all__ = ("CRTPDevice",)


class CRTPDevice:
    """Superclass for devices that use the CRTP protocol for communication."""

    _dispatcher: CRTPDispatcher
    _driver: CRTPDriver
    _exit_stack: Optional[AsyncExitStack]
    _task_group: Optional[DaemonTaskGroup]
    _uri: str

    def __init__(self, uri: str):
        """Constructor.

        Creates a CRTPDevice_ from a URI specification.

        Parameters:
            uri: the URI describing the data link where the device can be reached
        """
        self._uri = uri
        self._dispatcher = CRTPDispatcher()

        self._driver = None  # type: ignore
        self._task_group = None
        self._exit_stack = None

    async def __aenter__(self):
        assert self._exit_stack is None

        exit_stack = AsyncExitStack()
        await exit_stack.__aenter__()

        self._task_group = create_daemon_task_group()
        try:
            spawner = await exit_stack.enter_async_context(self._task_group)
            await spawner.start(self._open_connection)
            self._exit_stack = exit_stack
        finally:
            if self._exit_stack is None:
                await exit_stack.__aexit__(*exc_info())

        return self

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        assert self._exit_stack is not None
        exit_stack = self._exit_stack
        self._exit_stack = None
        return await exit_stack.__aexit__(exc_type, exc_value, tb)

    async def _message_handler(self, driver):
        """Worker task that receives incoming messages from the device and
        dispatches them to the coroutines that are interested in receiving these
        messages.

        Raises:
            IOError: when an IO error happens while receiving packets from the
                device (typically when the USB port or the radio dongle is
                disconnected)
        """
        while True:
            packet = await driver.receive_packet()
            await self._dispatcher.dispatch(packet)

    async def _open_connection(self, notify_started: TaskStartedNotifier):
        """Task that opens a connection to the device and yields control
        to the message handler task until the connection is closed.

        Parameters:
            notify_started: an async function that must be called and awaited
                when the connection has been established
        """
        async with CRTPDriver.connected_to(self._uri) as driver:
            self._driver = driver

            # TODO(ntamas): maybe handle IOError here?
            try:
                await self._prepare_link(driver)
                notify_started()
                await self._message_handler(driver)
            finally:
                self._driver = None  # type: ignore

    async def _prepare_link(self, driver: CRTPDriver) -> None:
        """Performs initial setup on the CRTP driver before the connection is
        declared to be established.
        """
        pass

    @property
    def dispatcher(self) -> CRTPDispatcher:
        """Returns the packet dispatcher that dispatches incoming messages to
        the appropriate handler functions.

        You may then use the dispatcher to register handler functions to the
        messages you are interested in.
        """
        return self._dispatcher

    async def packets(
        self, port: Optional[CRTPPortLike] = None, *, queue_size: int = 0
    ) -> AsyncIterable[CRTPPacket]:
        """Asynchronous generator that yields incoming packets matching the
        given port.

        Parameters:
            port: the CRTP port to match; `None` means to match all CRTP ports
            queue_size: number of pending packets that may stay in the backlog
                of the generator before blocking. Typically 0 is enough.

        Yields:
            incoming CRTP packets
        """
        async with self.dispatcher.create_packet_queue(
            port=port, queue_size=queue_size
        ) as queue:
            async for packet in queue:
                yield packet

    async def run_command(
        self,
        *,
        port: CRTPPortLike,
        channel: int = 0,
        command: Optional[CRTPCommandLike] = None,
        data: Optional[CRTPDataLike] = None,
        timeout: float = 0.2,
        attempts: int = 3,
    ) -> bytes:
        """Sends a command packet to the device and waits for the next
        matching response packet. Returns the data section of the response
        packet.

        The response packet is expected to have the same CRTP port and channel
        as the packet that was sent as a request. Additionally, if a command
        is present, the data section of the response packet is expected to
        start with the same command byte(s), and _only_ the rest of the data
        section is returned as the response. If there is no command specified,
        the response must match the port and the channel only, and the entire
        data section of the response packet will be returned.

        Parameters:
            port: the CRTP port to send the packet to
            channel: the CRTP channel to send the packet to
            command: the command byte(s) to insert before the data bytes in
                the data section of the packet. When this is not `None`, the
                matching response packet is expected to have the same prefix as
                the command itself.
            data: the data of the request packet. When a command is present, the
                command is inserted before the data in the body of the CRTP
                packet.
            timeout: maximum number of seconds to wait for a response
            attempts: maximum number of attempts to send the command and wait
                for the response

        Returns:
            the data section of the response packet
        """
        if command is not None:
            encoded_command = _handle_data_argument(command)
            packet = CRTPPacket(port=port, channel=channel)
            request = (encoded_command + bytes(data)) if data else encoded_command
            packet.data = request

            def matching_response(packet: CRTPPacket) -> bool:
                return packet.channel == channel and packet.data.startswith(
                    encoded_command
                )

        else:
            encoded_command = b""
            packet = CRTPPacket(port=port, channel=channel, data=data)

            def matching_response(packet: CRTPPacket) -> bool:
                return packet.channel == channel

        response = None

        # Send the packet and wait for the corresponding response
        with self.dispatcher.wait_for_next_packet(
            matching_response, port=port
        ) as value:
            while attempts > 0:
                # TODO(ntamas): self._driver may become `None` here if the
                # connection is closed for any reason in another async task.
                # This needs to be handled.
                await self._driver.send_packet(packet)
                if timeout > 0:
                    with move_on_after(timeout):
                        response = await value.wait()
                else:
                    response = await value.wait()

                # If there was no timeout (i.e. the response is not `None`),
                # break out.
                if response is not None:
                    break

                # Otherwise, try again.
                attempts -= 1

        if response is None:
            raise TimeoutError()

        response = response.data
        return response[len(encoded_command) :] if encoded_command else response

    async def send_packet(
        self,
        *,
        port: CRTPPortLike,
        channel: int = 0,
        data: Optional[Union[int, bytes, Iterable[Union[int, bytes]]]] = None,
    ):
        """Sends a packet to the device with the given CRTP port, channel and
        the given body.

        Parameters:
            port: the CRTP port to send the packet to
            channel: the CRTP channel to send the packet to
            data: the body of the request packet
        """
        packet = CRTPPacket(port=port, channel=channel)
        packet.data = _handle_data_argument(data)
        await self._driver.send_packet(packet)


def _handle_data_argument(command: Optional[CRTPCommandLike] = None) -> bytes:
    """Helper function to handle the conversion of the `command` argument in
    `CRTPDevice.run_command()` to a `bytes` object.
    """
    if command is None:
        return b""
    elif isinstance(command, int):
        return bytes((command,))
    elif isinstance(command, bytes):
        return command
    else:
        parts = []
        for part in command:
            if isinstance(part, int):
                part = bytes((part,))
            parts.append(part)
        return b"".join(parts)



"""Classes related to handling platform service messages of a Crazyflie."""

from enum import IntEnum
from typing import Any, Dict, Optional

from aiocflib.crtp import CRTPPort, LinkControlChannel

from .crazyflie import Crazyflie

__all__ = ("Platform",)


class PlatformChannel(IntEnum):
    """Enum representing the names of the channels of the platform service in
    the CRTP protocol.
    """

    PLATFORM_COMMAND = 0
    VERSION_COMMAND = 1
    APP_CHANNEL = 2


class PlatformCommand(IntEnum):
    """Enum representig the names of the platform commands in the platform
    service of the CRTP protocol.
    """

    SET_CONTINUOUS_WAVE = 0


class VersionCommand(IntEnum):
    """Enum representig the names of the version commands in the platform
    service of the CRTP protocol.
    """

    GET_PROTOCOL_VERSION = 0
    GET_FIRMWARE_VERSION = 1
    GET_DEVICE_TYPE_NAME = 2


class Platform:
    """Class representing the handler of platform service messages for a Crazyflie
    instance.
    """

    _crazyflie: Crazyflie
    _cache: Dict[str, Any]

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the platform
                service messages
        """
        self._crazyflie = crazyflie

        self._cache = {}
        # TODO(ntamas): clear cache when the Crazyflie disconnects

    async def get_device_type_name(self) -> Optional[str]:
        """Returns the device type name of the Crazyflie; `None` if the connected
        device is not a Crazyflie.
        """
        if "device_type_name" not in self._cache:
            self._cache["device_type_name"] = await self._get_device_type_name()
        return self._cache["device_type_name"]

    async def _get_device_type_name(self) -> Optional[str]:
        if await self.is_crazyflie():
            response = await self._crazyflie.run_command(
                port=CRTPPort.PLATFORM,
                channel=PlatformChannel.VERSION_COMMAND,
                command=VersionCommand.GET_DEVICE_TYPE_NAME,
            )
            return response.decode("ISO-8859-1", errors="backslashreplace")
        else:
            return None

    async def get_firmware_revision(self) -> Optional[str]:
        """Returns the firmware revision of the Crazyflie; `None` if the connected
        device is not a Crazyflie.
        """
        if "firmware_revision" not in self._cache:
            self._cache["firmware_revision"] = await self._get_firmware_revision()
        return self._cache["firmware_revision"]

    async def get_firmware_version(self) -> Optional[str]:
        """Returns the firmware version of the Crazyflie; `None` if the connected
        device is not a Crazyflie.
        """
        if "firmware_version" not in self._cache:
            self._cache["firmware_version"] = await self._get_firmware_version()
        return self._cache["firmware_version"]

    async def _get_firmware_revision(self) -> Optional[str]:
        if await self.is_crazyflie():
            rev0 = await self._crazyflie.param.get("firmware.revision0")
            rev1 = await self._crazyflie.param.get("firmware.revision1")
            return f"{rev0:08x}{rev1:04x}"
        else:
            return None

    async def _get_firmware_version(self) -> Optional[str]:
        if await self.is_crazyflie():
            response = await self._crazyflie.run_command(
                port=CRTPPort.PLATFORM,
                channel=PlatformChannel.VERSION_COMMAND,
                command=VersionCommand.GET_FIRMWARE_VERSION,
            )
            return response.decode("ISO-8859-1", errors="backslashreplace")
        else:
            return None

    async def get_protocol_version(self) -> int:
        """Returns the protocol version of the Crazyflie; -1 if the connected
        device is not a Crazyflie or it is too old and cannot supply the
        protocol version.
        """
        if "protocol_version" not in self._cache:
            self._cache["protocol_version"] = await self._get_protocol_version()
        return self._cache["protocol_version"]

    async def _get_protocol_version(self) -> int:
        # TODO(ntamas): cache locally, clear cache when the Crazyflie
        # disconnects
        if await self.is_crazyflie():
            response = await self._crazyflie.run_command(
                port=CRTPPort.PLATFORM,
                channel=PlatformChannel.VERSION_COMMAND,
                command=VersionCommand.GET_PROTOCOL_VERSION,
            )
            return response[0] if response else -1
        else:
            return -1

    async def is_crazyflie(self) -> bool:
        """Determines whether the connected device is a Crazyflie.

        Returns:
            whether the connected device is a Crazyflie
        """
        if "is_crazyflie" not in self._cache:
            self._cache["is_crazyflie"] = await self._is_crazyflie()
        return self._cache["is_crazyflie"]

    async def _is_crazyflie(self) -> bool:
        response = await self._crazyflie.run_command(
            port=CRTPPort.LINK_CONTROL, channel=LinkControlChannel.SOURCE, data=b"\x00"
        )
        return response.startswith(b"Bitcraze Crazyflie")

    async def packets(self):
        """Async generator that yields platform service messages from a
        Crazyflie.
        """
        async for packet in self._crazyflie.packets(port=CRTPPort.PLATFORM):
            yield packet



"""Functions related to rotations and quaternions."""

from math import sqrt
from typing import List, NamedTuple

QuaternionXYZW = NamedTuple(
    "QuaternionXYZW", [("x", float), ("y", float), ("z", float), ("w", float)]
)

SQRT1_2 = 1.0 / sqrt(2)


def normalize_quaternion(quat: QuaternionXYZW) -> QuaternionXYZW:
    """Normalizes a quaternion to unit length."""
    length = sqrt(sum(x**2 for x in quat))
    x, y, z, w = quat
    return QuaternionXYZW(x / length, y / length, z / length, w / length)


def compress_unit_quaternion(quat: QuaternionXYZW, *, normalize: bool = False) -> int:
    """Compresses a generic XYZW quaternion of unit length into a 32-bit
    unsigned integer representation.

    In the compressed representation, we make use of the following:

    - we only send the smallest three components as the fourth component can be
      inferred for unit quaternions

    - the second-largest element can be at most 1 / sqrt(2) so we can normalize
      with that again to increase resolution

    - we assume that the sign of the largest component is positive so we don't
      need to encode that

    - we use 9 bits plus 1 sign bit per component

    Parameters:
        quat: the quaternion to compress
        normalize: whether to normalize the quaternion before compressing it.
            Setting this variable to ``False`` means that the caller _knows_
            that the quaternion is already normalized, and the function will
            fail if it is not normalized.

    Returns:
        the compressed representation of the quaternion
    """
    if normalize:
        quat = normalize_quaternion(quat)

    largest_index = 0
    for index in range(4):
        if abs(quat[index]) > abs(quat[largest_index]):
            largest_index = index

    negate = quat[largest_index] < 0
    comp = largest_index
    for index in range(4):
        if index != largest_index:
            negbit = (quat[index] < 0) ^ negate
            mag = int(round(((1 << 9) - 1) * (abs(quat[index]) / SQRT1_2)))
            comp = (comp << 10) | (negbit << 9) | mag

    return comp


def decompress_unit_quaternion(quat_compressed: int) -> QuaternionXYZW:
    """Decompresses a generic XYZW quaternion of unit length from its 32-bit
    unsigned integer representation.

    See `compress_unit_quaternion()` for more details.

    Parameters:
        quat_compressed: the compressed quaternion to decompress

    Returns:
        the standard representation of the quaternion
    """
    mask: int = (1 << 9) - 1
    largest_index: int = (quat_compressed >> 30) & 0x3
    sum_squares: float = 0.0
    result: List[float] = [0.0] * 4

    for index in range(3, -1, -1):
        if index != largest_index:
            mag = quat_compressed & mask
            negate = (quat_compressed >> 9) & 1
            quat_compressed >>= 10
            result[index] = (SQRT1_2 * mag) / mask
            if negate:
                result[index] *= -1
            sum_squares += result[index] * result[index]

    result[largest_index] = sqrt(1.0 - sum_squares)
    return QuaternionXYZW(*result)



"""Classes and functions related to the handling and caching of parameter
and log table-of-contents entries from a Crazyflie.
"""

from abc import abstractmethod, ABCMeta
from anyio import Lock, open_file
from binascii import hexlify
from collections import defaultdict
from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from struct import Struct
from typing import Awaitable, Callable, Iterable, Optional, Union, Tuple, TypeVar

from aiocflib.utils.registry import Registry

__all__ = ("TOCCache",)


#: Type alias for items stored in a TOC cache
TOCItem = bytes

#: Type alias for TOC namespaces
Namespace = str

#: Type alias for objects from which we can create a TOC cache instance
TOCCacheLike = Union[None, str, Path, "TOCCache"]


class TOCCache(metaclass=ABCMeta):
    """Interface specification for table-of-contents caches."""

    @classmethod
    def create(cls, spec: TOCCacheLike):
        """Creates a table-of-contents cache from a URI-style string
        specification or a filesystem path.

        Parameters:
            spec: a Python Path_ object pointing to a folder in the
                filesystem, a URI-style cache specification, or a string
                pointing to a folder in the filesystem. `None` means to
                create a null cache that does nothing.
        """
        if spec is None:
            spec = "null://"

        if isinstance(spec, TOCCache):
            return spec
        elif isinstance(spec, Path):
            return cls.create(str(spec))
        else:
            scheme, sep, rest = spec.partition("://")
            if not sep:
                scheme, rest = "file", scheme

            try:
                factory = TOCCacheRegistry.find(scheme)
            except KeyError:
                raise KeyError("no such TOC cache type: {0!r}".format(scheme)) from None

            cache = factory()
            cache._configure(rest)

        return cache

    def _configure(self, uri: str) -> None:  # noqa: B027
        """Configures the cache instance from the given URI specification.

        This method is not public; use the `TOCCache.create()` constructor to
        create a TOC cache from a URI.

        Since the method is not meant for public use, it does _not_ check
        whether the received URI has a scheme that corresponds to the TOC cache
        type; it is assumed that the caller already took care of it.

        Parameters:
            uri: the URI to create the cache instance from
        """
        pass

    async def find(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Iterable[TOCItem]:
        """Looks up a cached table-of-contents by its hash value.

        Parameters:
            hash: the hash code to look up
            namespace: the namespace that the hash value should be looked up
                in. Useful if the same TOC cache is used to store different
                types of TOC items.

        Raises:
            KeyError: if no cached table-of-contents entries exist for the
                given hash value
        """
        raise NotImplementedError

    async def has(self, hash: bytes, namespace: Optional[Namespace] = None) -> bool:
        """Returns whether the table-of-contents object contains an item with
        the given hash.

        Parameters:
            hash: the hash code to look up
            namespace: the namespace that the hash value should be looked up
                in. Useful if the same TOC cache is used to store different
                types of TOC items.
        """
        raise NotImplementedError

    def get_key(self) -> Optional[str]:
        """Returns a short string identifier that uniquely identifies this
        cache, or `None` if no such key can be derived.

        The purpose of this method is to allow us to create locks to
        semantically equivalent caches. The idea is that the locks are
        associated to the keys of the TOC caches; in other words, if two
        caches have the same key, then locking one of them implicitly locks the
        other and vice versa.
        """
        return None

    def namespace(self, namespace: str) -> "TOCCache":
        """Returns another TOC cache instance that is restricted to the given
        namespace.

        Retrieving or setting a hash in the base namespace of the returned
        instance will retrieve or set it in the given namespace of the wrapped
        instance.
        """
        return NamespacedTOCCacheWrapper(self, namespace)

    @abstractmethod
    async def store(
        self,
        hash: bytes,
        items: Iterable[TOCItem],
        namespace: Optional[Namespace] = None,
    ) -> None:
        """Stores a list of table-of-contents entries under the given hash code.

        Parameters:
            hash: the hash code to use up
            items: the entries in the table-of-contents object to store
            namespace: the namespace that the hash value should be looked up
                in. Useful if the same TOC cache is used to store different
                types of TOC items.

        """
        raise NotImplementedError


#: Type alias for factory functions that can create a TOC cache instance
TOCCacheFactory = Callable[[], TOCCache]

#: Mapping that maps names to the corresponding TOCCache classes
TOCCacheRegistry = Registry()  # type: Registry[TOCCacheFactory]


@TOCCacheRegistry.register("null")
class NullTOCCache(TOCCache):
    """Null TOC cache that does nothing."""

    async def find(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Iterable[TOCItem]:
        raise KeyError("no such hash: {0!r}".format(hash))

    async def has(self, hash: bytes, namespace: Optional[Namespace] = None) -> bool:
        return False

    def get_key(self) -> Optional[str]:
        return "null"

    async def store(
        self,
        hash: bytes,
        items: Iterable[TOCItem],
        namespace: Optional[Namespace] = None,
    ) -> None:
        pass


@TOCCacheRegistry.register("memory")
class InMemoryTOCCache(TOCCache):
    """TOC cache specialization that stores the table-of-contents entries
    purely in memory.
    """

    def __init__(self):
        """Constructor."""
        self._namespaces = defaultdict(dict)

    async def find(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Iterable[TOCItem]:
        items = self._namespaces.get(namespace)
        if not items:
            raise KeyError("no such namespace: {0!r}".format(namespace))

        try:
            return items[hash]
        except KeyError:
            raise KeyError("no such hash: {0!r}".format(hash)) from None

    async def has(self, hash: bytes, namespace: Optional[Namespace] = None) -> bool:
        try:
            await self.find(hash, namespace)
            return True
        except KeyError:
            return False

    async def store(
        self,
        hash: bytes,
        items: Iterable[TOCItem],
        namespace: Optional[Namespace] = None,
    ) -> None:
        self._namespaces[namespace][hash] = tuple(items)


@TOCCacheRegistry.register("file")
class FilesystemBasedTOCCache(TOCCache):
    """TOC cache specialization that stores the table-of-contents entries on
    a filesystem in a given folder.
    """

    _key: Optional[Path]
    _read_only: bool
    _path: Optional[Path]

    def __init__(self, read_only: bool = False):
        """Constructor.

        Parameters:
            read_only: whether the cache is read only
        """
        self._path = None
        self._read_only = bool(read_only)
        self._key = None

    def _configure(self, uri: str) -> None:
        self.path = uri

    def get_key(self) -> Optional[str]:
        if self._key is None:
            self._key = self.path.resolve() if self.path else None
        return str(self._key)

    @property
    def path(self) -> Path:
        if self._path is None:
            raise ValueError("TOC cache is not configured yet")
        return self._path

    @path.setter
    def path(self, value) -> None:
        if self._path is not None:
            raise ValueError("TOC cache is already configured")

        self._path = Path(value)
        self._key = None

    def _path_for_hash(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Path:
        """Sanitizes the given hash so it can be used in a filesystem path."""
        path = self._path_for_namespace(namespace)
        return path / "{0}.bin".format(hexlify(hash).decode("ascii").lower())

    def _path_for_namespace(self, namespace: Optional[Namespace] = None) -> Path:
        """Returns the path corresponding to the given namespace on the filesystem."""
        if self._path is None:
            raise ValueError("TOC cache is not configured yet")

        result = self._path
        if namespace is not None:
            namespace = self._sanitize_namespace(namespace)

        return result / namespace if namespace else result

    @staticmethod
    def _sanitize_namespace(namespace: Namespace) -> str:
        """Sanitizes the given namespace so it can be used in a filesystem
        path.
        """
        if namespace is None:
            return ""
        return (
            namespace.encode("ascii", errors="backslashreplace")
            .replace(b"/", b"\\2f")
            .replace(b"\\", b"=")
            .replace(b".", b"/")
            .decode("ascii")
        )

    async def find(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Iterable[TOCItem]:
        path = self._path_for_hash(hash, namespace)
        if not path.exists() or not path.is_file():
            raise KeyError(
                "no such namespace or hash: {0!r} / {1!r}".format(namespace, hash)
            )

        result = []

        async with await open_file(str(path), "rb") as fp:
            data = await fp.read(1)
            if not data:
                raise IOError("unexpected end of file")

            version = data[0]
            if version != 1:
                raise IOError("only version 1 TOC files are supported")

            while True:
                length = await fp.read(2)
                if length is None or len(length) < 2:
                    break

                assert isinstance(length, bytes)
                length = length[0] + (length[1] << 8)
                data = await fp.read(length)
                if len(data) < length:
                    raise IOError("unexpected end of file")

                result.append(data)

        return result

    async def has(self, hash: bytes, namespace: Optional[Namespace] = None) -> bool:
        path = self._path_for_hash(hash, namespace)
        return path.exists() and path.is_file()

    async def store(
        self,
        hash: bytes,
        items: Iterable[TOCItem],
        namespace: Optional[Namespace] = None,
    ) -> None:
        if self._read_only:
            return

        path = self._path_for_hash(hash, namespace)
        path.parent.mkdir(parents=True, exist_ok=True)

        success = False

        try:
            async with await open_file(str(path), "wb") as fp:
                await fp.write(b"\x01")
                for item in items:
                    length = len(item)
                    if length > 65535:
                        raise IOError("exceeded maximum item length")

                    await fp.write(bytes((length & 0xFF, length >> 8)))
                    await fp.write(item)

            success = True
        finally:
            if not success:
                path.unlink()


TOCCacheRegistry.register("file+ro", partial(FilesystemBasedTOCCache, read_only=True))


class NamespacedTOCCacheWrapper(TOCCache):
    """TOCCache implementation that wraps another TOC cache and ensures that
    the operations are performed only in a given namespace of the wrapped
    instance.
    """

    def __init__(self, wrapped: TOCCache, namespace: Namespace, sep: str = "."):
        self._wrapped = wrapped
        self._namespace = namespace
        self._separator = sep

    async def find(
        self, hash: bytes, namespace: Optional[Namespace] = None
    ) -> Iterable[TOCItem]:
        return await self._wrapped.find(hash, self._remap(namespace))

    def get_key(self) -> Optional[str]:
        if self._wrapped is not None and self._namespace:
            root_key = self._wrapped.get_key()
            return f"{root_key}{self._separator}{self._namespace}"
        else:
            return None

    async def has(self, hash: bytes, namespace: Optional[Namespace] = None) -> bool:
        return await self._wrapped.has(hash, self._remap(namespace))

    async def store(
        self,
        hash: bytes,
        items: Iterable[TOCItem],
        namespace: Optional[Namespace] = None,
    ) -> None:
        return await self._wrapped.store(hash, items, self._remap(namespace))

    def _remap(self, namespace: Optional[Namespace]) -> Namespace:
        if namespace is None:
            return self._namespace
        else:
            return self._separator.join((self._namespace, namespace))


#: Dictionary that maps TOCCache classes to dictionaries that map cache keys to their locks
_cache_locks = defaultdict(lambda: defaultdict(Lock))


@asynccontextmanager
async def _locked_cache(cache: Optional[TOCCache], key_suffix: Optional[str] = None):
    key = cache.get_key() if cache is not None else None
    if key is None:
        yield
    else:
        cls = cache.__class__
        locks = _cache_locks[cls]
        if key_suffix:
            key = f"{key}:{key_suffix}"
        async with locks[key]:
            yield


T = TypeVar("T")


async def fetch_table_of_contents_gracefully(
    cache: Optional[TOCCache],
    info_func: Callable[[], Awaitable[Tuple[int, int]]],
    single_item_fetcher_func: Callable[[int], Awaitable[T]],
    from_bytes: Callable[[bytes, int], T],
    to_bytes: Callable[[T], bytes],
):
    num_items, hash = await info_func()
    hash = Struct("<I").pack(hash)
    result = None

    async with _locked_cache(cache, key_suffix=hash.hex()):
        try:
            # Try to fetch the parameters from the cache based on the hash
            if cache:
                items = await cache.find(hash)
                result = [from_bytes(data, id) for id, data in enumerate(items)]
        except Exception:
            pass

        if result is None:
            # Retrieving the cached entries failed; let's try to fetch on
            # our own
            result = [await single_item_fetcher_func(i) for i in range(num_items)]

            # Store the fetched entries in the cache
            if cache:
                try:
                    await cache.store(hash, [to_bytes(item) for item in result])
                except Exception:
                    # Storing items in the cache failed, but let's not freak out
                    pass

    return result



"""Classes related to sending high-level navigation commands to a Crazyflie.
"""

from contextlib import asynccontextmanager
from enum import IntEnum
from math import radians
from struct import Struct
from typing import AsyncIterator, ClassVar, Optional

from aiocflib.crtp import CRTPCommandLike, CRTPDataLike, CRTPPort
from aiocflib.errors import CRTPCommandError

from .crazyflie import Crazyflie

__all__ = ("HighLevelCommander", "HighLevelCommanderError")


class HighLevelCommanderError(CRTPCommandError):
    """CRTPCommandError subclass for errors emitted from the high-level
    commander module.
    """

    pass


class HighLevelCommand(IntEnum):
    """Enum representing the names of the high-level navigation commands in the
    high-level commander service of the CRTP protocol.
    """

    SET_GROUP_MASK = 0
    TAKEOFF = 1
    LAND = 2
    STOP = 3
    GO_TO = 4
    START_TRAJECTORY = 5
    DEFINE_TRAJECTORY = 6
    TAKEOFF_2 = 7
    LAND_2 = 8
    TAKEOFF_WITH_VELOCITY = 9
    LAND_WITH_VELOCITY = 10


class TrajectoryType(IntEnum):
    """Enum representing the possible trajectory types supported by the
    Crazyflie.
    """

    POLY4D = 0
    COMPRESSED = 1


class TrajectoryLocation(IntEnum):
    """Enum representing the possible locations where a trajectory can reside
    on the Crazyflie.
    """

    INVALID = 0
    MEMORY = 1


#: Default group mask representing "all groups"
ALL_GROUPS = 0


class HighLevelCommander:
    """Class responsible for sending high-level navigation commands to a
    Crazyflie.
    """

    _crazyflie: Crazyflie

    _define_trajectory_struct: ClassVar[Struct] = Struct("<BBIB")
    _go_to_struct: ClassVar[Struct] = Struct("<Bfffff")
    _land_struct: ClassVar[Struct] = Struct("<ff?f")
    _land_with_velocity_struct: ClassVar[Struct] = Struct("<f?f?f")
    _start_trajectory_struct: ClassVar[Struct] = Struct("<BBBf")
    _takeoff_struct: ClassVar[Struct] = Struct("<ff?f")
    _takeoff_with_velocity_struct: ClassVar[Struct] = Struct("<f?f?f")

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie to which we need to send the messages
        """
        self._crazyflie = crazyflie

    async def define_trajectory(
        self,
        id: int,
        *,
        addr: int,
        num_pieces: int = 0,
        location: TrajectoryLocation = TrajectoryLocation.MEMORY,
        type: TrajectoryType = TrajectoryType.POLY4D,
    ) -> None:
        """Defines a trajectory in the trajectory memory of the Crazyflie.

        Parameters:
            id: ID of the trajectory to define
            addr: address where the trajectory starts in the trajectory memory
            num_pieces: number of segments in the trajectory if it is stored in
                uncompressed format; ignored for compressed trajectories
            location: specifies where (in which memory) the trajectory is on the
                Crazyflie
            type: specifies the type (encoding) of the trajectory
        """
        data = self._define_trajectory_struct.pack(location, type, addr, num_pieces)
        await self._run_command(
            command=(HighLevelCommand.DEFINE_TRAJECTORY, id), data=data
        )

    async def disable(self) -> None:
        """Disables the high-level controller on the Crazyflie."""
        await self._crazyflie.parameters.set("commander.enHighLevel", 0)

    async def enable(self) -> None:
        """Enables the high-level controller on the Crazyflie."""
        await self._crazyflie.parameters.set("commander.enHighLevel", 1)

    @asynccontextmanager
    async def enabled(self) -> AsyncIterator[None]:
        """Async context manager that enables the high-level controller when
        entering the context and disables it when exiting the context.
        """
        async with self._crazyflie.parameters.set_and_restore(
            "commander.enHighLevel", 1, 0
        ):
            yield

    async def go_to(
        self,
        x: float,
        y: float,
        z: float,
        *,
        duration: float,
        yaw: float = 0.0,
        relative: bool = False,
        group_mask: int = ALL_GROUPS,
    ) -> None:
        """Sends a command to navigate to the given absolute or relative
        position.

        Parameters:
            x: the X coordinate of the new position
            y: the Y coordinate of the new position
            z: the Z coordinate of the new position
            yaw: the yaw angle to reach at the new position, in degrees
            duration: duration of the movement, in seconds
            relative: defines whether the coordinates are absolute (`False`)
                or relative to the current position (`True`)
            group_mask: mask that defines which Crazyflie drones this command
                should apply to
        """
        data = self._go_to_struct.pack(
            relative,
            x,
            y,
            z,
            radians(yaw),
            duration,
        )
        await self._run_command(command=(HighLevelCommand.GO_TO, group_mask), data=data)

    async def is_enabled(self, fetch: bool = False) -> bool:
        """Retrieves whether the high-level command is currently enabled on
        the Crazyflie.

        Parameters:
            fetch: whether to forcefully fetch the current value of this
                parameter from the drone even if we have a locally cached copy
        """
        value = await self._crazyflie.parameters.get(
            "commander.enHighLevel", fetch=fetch
        )
        return bool(value)

    async def land(
        self,
        height: float,
        *,
        duration: Optional[float] = None,
        velocity: Optional[float] = None,
        relative: Optional[bool] = False,
        yaw: Optional[float] = None,
        group_mask: int = ALL_GROUPS,
    ) -> None:
        """Sends a landing command to the Crazyflie.

        The landing velocity may be specified directly (with the `velocity`
        parameter) or indirectly (with the `duration` parameter). When neither
        the velocity nor the duration are prescribed, a default safe landing
        velocity will be applied; this is decided by the firmware.

        Parameters:
            height: the absolute or relative height to land to, in meters
            duration: duration of the landing, in seconds; mutually exclusive
                with `velocity`
            velocity: average velocity of the landing, in m/s; mutually
                exclusive with `duration`
            relative: whether the landing height is relative to the current
                height; positive relative height is below the current height
            yaw: the target yaw, in degrees; `None` to use the current yaw
            group_mask: mask that defines which Crazyflie drones this command
                should apply to
        """
        if duration is not None and velocity is not None:
            raise ValueError("duration and velocity are mutually exclusive")

        if duration is not None:
            if duration < 0:
                raise ValueError("duration may not be negative")
            elif duration == 0:
                duration = None

        use_current_yaw = yaw is None
        yaw = yaw or 0.0

        if relative and duration is not None:
            # Firmware supports relative height only when the velocity is
            # specified, so we need to turn the duration into velocity
            velocity, duration = abs(height) / duration, None

        if velocity is None:
            velocity = 0  # firmware picks a velocity
        elif velocity < 0:
            raise ValueError("velocity may not be negative")

        # Convert yaw into radians
        yaw = radians(yaw)

        # Decide which command to use
        if duration is not None:
            data = self._land_struct.pack(height, yaw, use_current_yaw, duration)
            await self._run_command(
                command=(HighLevelCommand.LAND_2, group_mask), data=data
            )
        else:
            data = self._land_with_velocity_struct.pack(
                height, relative, yaw, use_current_yaw, velocity
            )
            await self._run_command(
                command=(HighLevelCommand.LAND_WITH_VELOCITY, group_mask), data=data
            )

    async def set_group_mask(self, group_mask: int = ALL_GROUPS) -> None:
        """Sets the group mask of the Crazyflie, defining which groups the
        drone will belong to.

        Parameters:
            group_mask: mask that defines which groups this Crazyflie drone
                belongs to
        """
        await self._run_command(command=(HighLevelCommand.SET_GROUP_MASK, group_mask))

    async def start_trajectory(
        self,
        id: int,
        *,
        time_scale: float = 1,
        relative: bool = False,
        reversed: bool = False,
        group_mask: int = ALL_GROUPS,
    ) -> None:
        """Starts a trajectory that was already uploaded to the trajectory memory
        of the Crazyflie.

        Parameters:
            id: ID of the trajectory to start
            time_scale: specifies the time scale of the playback; 1.0 means
                "native speed" (as prescribed by the trajectory), >1.0 slows
                playback down, <1.0 speeds playback up
            relative: whether to treat the trajectory coordinates relative to
                the current position of the Crazyflie
            reversed: whether to play the trajectory backwards. Not supported
                for compressed trajectories
        """
        data = self._start_trajectory_struct.pack(
            relative,
            reversed,
            id,
            time_scale,
        )
        await self._run_command(
            command=(HighLevelCommand.START_TRAJECTORY, group_mask), data=data
        )

    async def stop(self, group_mask: int = ALL_GROUPS) -> None:
        """Sends a command to the Crazyflie to turn off the motors immediately.

        Parameters:
            group_mask: mask that defines which Crazyflie drones this command
                should apply to
        """
        await self._run_command(command=(HighLevelCommand.STOP, group_mask))

    async def takeoff(
        self,
        height: float,
        *,
        duration: Optional[float] = None,
        velocity: Optional[float] = None,
        relative: Optional[bool] = False,
        yaw: Optional[float] = None,
        group_mask: int = ALL_GROUPS,
    ) -> None:
        """Sends a takeoff command to the Crazyflie.

        The takeoff velocity may be specified directly (with the `velocity`
        parameter) or indirectly (with the `duration` parameter). When neither
        the velocity nor the duration are prescribed, a default safe takeoff
        velocity will be applied; this is decided by the firmware.

        Parameters:
            height: the absolute or relative height to take off to, in meters
            duration: duration of the takeoff, in seconds; mutually exclusive
                with `velocity`
            velocity: average velocity of the takeoff, in m/s; mutually
                exclusive with `duration`
            relative: whether the takeoff height is relative to the current
                height; positive relative height is above the current height
            yaw: the target yaw, in degrees; `None` to use the current yaw
            group_mask: mask that defines which Crazyflie drones this command
                should apply to
        """
        if duration is not None and velocity is not None:
            raise ValueError("duration and velocity are mutually exclusive")
        if duration is not None:
            if duration < 0:
                raise ValueError("duration may not be negative")
            elif duration == 0:
                duration = None

        use_current_yaw = yaw is None
        yaw = yaw or 0.0

        if relative and duration is not None:
            # Firmware supports relative height only when the velocity is
            # specified, so we need to turn the duration into velocity
            velocity, duration = abs(height) / duration, None

        if velocity is None:
            velocity = 0  # firmware picks a velocity
        elif velocity < 0:
            raise ValueError("velocity may not be negative")

        # Convert yaw into radians
        yaw = radians(yaw)

        # Decide which command to use
        if duration is not None:
            data = self._takeoff_struct.pack(height, yaw, use_current_yaw, duration)
            await self._run_command(
                command=(HighLevelCommand.TAKEOFF_2, group_mask), data=data
            )
        else:
            data = self._takeoff_with_velocity_struct.pack(
                height, relative, yaw, use_current_yaw, velocity
            )
            await self._run_command(
                command=(HighLevelCommand.TAKEOFF_WITH_VELOCITY, group_mask), data=data
            )

    async def _run_command(
        self,
        *,
        command: Optional[CRTPCommandLike] = None,
        data: Optional[CRTPDataLike] = None,
        **kwds,
    ) -> None:
        """Sends a command packet to the high-level commander port and channel
        of the Crazyflie, waits for the next matching response packet and
        handles the error code in the packet.

        Keyword arguments not mentioned here are forwarded to the
        ``run_command()`` method of the Crazyflie itself.

        Parameters:
            command: the command byte(s) to insert before the data bytes in
                the data section of the packet. When this is not `None`, the
                matching response packet is expected to have the same prefix as
                the command itself.
            data: the data of the request packet. When a command is present, the
                command is inserted before the data in the body of the CRTP
                packet.

        Raises:
            HighLevelCommanderError: if the high-level commander refused to
                execute the command
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.HIGH_LEVEL_COMMANDER, command=command, data=data, **kwds
        )
        if len(response) < 1:
            raise HighLevelCommanderError(message="Response too short")
        elif response[-1]:
            raise HighLevelCommanderError(code=response[-1])



"""Utility functions related to concurrency management."""

from anyio import (
    CapacityLimiter,
    Condition,
    Event,
    create_task_group,
    from_thread,
    to_thread,
)
from anyio.abc import TaskGroup
from functools import partial
from inspect import iscoroutinefunction
from outcome import capture
from queue import Full, Queue
from sys import exc_info
from typing import (
    cast,
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

__all__ = (
    "AwaitableValue",
    "aclosing",
    "create_daemon_task_group",
    "Full",
    "ObservableValue",
    "ThreadContext",
)

T = TypeVar("T")

TaskStartedNotifier = Callable[[], None]


class aclosing:
    """Context manager that handles the closing of an asynchronous generator
    when the context is exited.
    """

    def __init__(self, aiter):
        self._aiter = aiter

    async def __aenter__(self):
        return self._aiter

    async def __aexit__(self, *args) -> bool:
        await self._aiter.aclose()
        return False


class AwaitableValue(Generic[T]):
    """Object that combines an asyncio-style event and a value. Initially, the
    object contains no value and the corresponding event is not set. Tasks may
    wait for the object to be populated with a value.
    """

    _event: Event
    _value: Optional[T]

    def __init__(self):
        """Constructor."""
        self._event = Event()
        self._value = None

    def set(self, value: T) -> None:
        """Sets a value and notifies all tasks waiting for the value."""
        if self._event.is_set():
            raise RuntimeError("awaitable value is already set")

        self._value = value
        self._event.set()

    async def wait(self) -> T:
        """Waits for the value to be populated, and returns the value when
        it is populated.
        """
        await self._event.wait()
        return self._value  # type: ignore


class DaemonTaskGroup(TaskGroup):
    """Task group that cancels all its child tasks when the execution is about
    to leave the context (instead of waiting for the child tasks to finish).
    """

    _task_group: TaskGroup

    def __init__(self):
        self._task_group = create_task_group()
        self._spawner = None

    async def __aenter__(self):
        self._spawner = await self._task_group.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        self._spawner = None

        await self._task_group.cancel_scope.cancel()
        return bool(await self._task_group.__aexit__(exc_type, exc_value, tb))

    def start_soon(self, func, *args, **kwds):
        if self._spawner is None:
            raise RuntimeError("DaemonTaskGroup is not running yet")
        return self._spawner.start_soon(func, *args, **kwds)

    async def start(self, func, *args, **kwds):
        on_opened = Event()
        self.start_soon(partial(func, notify_started=on_opened.set), *args, **kwds)
        await on_opened.wait()


create_daemon_task_group = DaemonTaskGroup


async def _gather_execute(
    func: Callable[..., Awaitable[T]], args: Any, result: List[T], index: int
) -> None:
    result[index] = await func(*args)


async def _gather_execute_limited(
    limiter: CapacityLimiter,
    func: Callable[..., Awaitable[T]],
    args: Any,
    result: List[T],
    index: int,
):
    async with limiter:
        result[index] = await func(*args)


async def gather(
    funcs: Iterable[Union[Callable[[], T], Tuple[Callable[..., T], ...]]],
    limiter: Optional[Union[CapacityLimiter, int]] = None,
) -> List[T]:
    to_execute = [
        (func, ()) if callable(func) else (func[0], func[1:]) for func in funcs
    ]
    result: List[Optional[T]] = []

    if isinstance(limiter, int):
        limiter = CapacityLimiter(limiter)

    run: Callable[..., Awaitable[Any]] = (
        _gather_execute
        if limiter is None
        else partial(_gather_execute_limited, limiter)
    )

    async with create_task_group() as group:
        for func, args in to_execute:
            if iscoroutinefunction(func):
                result.append(None)
                group.start_soon(run, func, args, result, len(result) - 1)
            else:
                result.append(func(*args))

    # At this point all None instances from result should be gone
    return cast(List[T], result)


class ObservableValue(Generic[T]):
    """Object that combines an asyncio-style condition and a value. Tasks may
    observe value changes in the object either by waiting for the value to
    change using the `wait()` method, or by starting a generator that yields
    new values by using the ObservableValue_ in an `async for` loop.

    Note that observers of the value are not guaranteed to receive all values
    if the value is changing rapidly. Whenever the generator running the observer
    task gets waken up, it retrieves the _current_ value and yields that to the
    caller.
    """

    @classmethod
    def constant(cls, value: T):
        return cls(value)

    def __init__(self, value: T):
        """Constructor.

        Parameters:
            value: the initial value
        """
        self._condition = Condition()
        self._value = value

    async def set(self, value: T, force: bool = False) -> None:
        """Sets a new value and notifies all tasks waiting for the value.

        Does not notify other tasks if the new value is exactly the same as
        the old one, unless `force` is set to `True`.
        """
        if value == self._value and not force:
            return

        self._value = value

        async with self._condition:
            self._condition.notify_all()

    update = set

    @property
    def value(self) -> T:
        """Returns the current value of the observable."""
        return self._value

    async def wait(self) -> T:
        """Waits for the value to change, and returns the most recent value
        at the earliest possible occasion.
        """
        async with self._condition:
            await self._condition.wait()
            return self._value

    async def wait_for(self, expected: T) -> T:
        """Blocks until the value becomes equal to an expected value."""
        async with self._condition:
            while True:
                if self._value == expected:
                    return self._value

                await self._condition.wait()

    async def wait_until(self, predicate: Callable[[T], bool]) -> T:
        """Blocks until the value satisfies a predicate."""
        async with self._condition:
            while True:
                if predicate(self._value):
                    return self._value

                await self._condition.wait()

    def __aiter__(self):
        return self._observe()

    async def _observe(self):
        last_value = self._value
        yield last_value

        while True:
            if last_value != self._value:
                last_value = self._value
                yield last_value
                # By the time we get back here, the value might have changed
                # again so we cannot call wait() now, we need to loop and
                # compare the current value with the last one
            else:
                await self.wait()


class ThreadContext(Generic[T]):
    """Context manager that spawns a thread when entering the context and
    kills the thread upon exiting the context.
    """

    #: Type alias for the target function of a ThreadContext
    Target = Callable[[Queue, Callable[[Any], None]], None]

    _queue: Optional[Queue]
    _task_group: Optional[TaskGroup]
    _value: Optional[AwaitableValue[T]]

    @classmethod
    def create_reader(
        cls, reader, queue_to_caller, *, setup=None, teardown=None, skip=None, **kwds
    ):
        """Convenience constructor for a common use-case: the thread is
        executing a blocking reader function in an infinite loop and sends the
        return values of the function in a queue back to the caller.

        Parameter:
            reader: the reader function to call periodically in the reader
                thread
            queue_to_caller: the queue in which the return values of the reader
                function will be placed
            setup: an optional setup function that will be called with no
                arguments _before_ the reader thread enters its main loop.
            teardown: an optional teardown function that will be called with
                an exception type, value and the associated traceback when the
                thread exits its main loop. The exception type, value and
                traceback are all `None` if the main loop terminated without
                an exception.
            skip: when specified, return values from the reader function that
                are equal to this value (using the ``is`` operator) will not be
                put into the queue that leads back to the caller.

        Additional keyword arguments are forwarded intact to the ThreadContext_
        constructor.
        """

        def respond_from_reader(value: Any) -> None:
            from_thread.run(queue_to_caller.send, value)

        def reader_thread(queue: Queue, on_started: Callable[[Any], None]) -> None:
            if setup:
                setup()

            on_started(queue_to_caller.receive)

            try:
                while True:
                    if not queue.empty():
                        item = queue.get()
                        if item is None:
                            break

                    value = reader()
                    if value is not skip:
                        respond_from_reader(value)

            finally:
                if teardown:
                    teardown(*exc_info())

        return cls(target=reader_thread, **kwds)

    @classmethod
    def create_worker(cls, *, setup=None, teardown=None, **kwds):
        """Convenience constructor for a common use-case: the thread is
        executing an infinite loop that receives functions to call via a queue,
        executes the functions sequentially, and passes the return values
        back to the caller in an AwaitableValue_.

        Parameter:
            setup: an optional setup function that will be called with no
                arguments _before_ the worker thread enters its main loop.
            teardown: an optional teardown function that will be called with
                an exception type, value and the associated traceback when the
                thread exits its main loop. The exception type, value and
                traceback are all `None` if the main loop terminated without
                an exception.

        Additional keyword arguments are forwarded intact to the ThreadContext_
        constructor.
        """

        def respond_from_worker(container: AwaitableValue, value: Any) -> None:
            from_thread.run_sync(container.set, value)

        def worker_thread(queue: Queue, on_started: Callable[[Any], None]):
            if setup:
                setup()

            sender = queue.put_nowait
            closed = False

            async def call_in_worker(func, *args, **kwargs):
                """Helper function that calls the given function in the worker
                thread with the given positional and keyword arguments.
                """
                if closed:
                    # This may happen if the main event loop has already sent
                    # us a request to terminate, but other tasks in the main
                    # event loop managed to sneak in some more requests into
                    # the queue. This should be handled better; for instance,
                    # these tasks should not be allowed to send requests once
                    # the queue is closed.
                    raise RuntimeError("Worker thread already closed")

                value = AwaitableValue()
                sender((func, args, kwargs, partial(respond_from_worker, value)))
                outcome = await value.wait()
                return outcome.unwrap()

            on_started(call_in_worker)

            try:
                while True:
                    item = queue.get()
                    if item is None:
                        closed = True
                        break

                    func, args, kwargs, responder = item
                    result = capture(func, *args, **kwargs)
                    if responder:
                        responder(result)

            finally:
                if teardown:
                    teardown(*exc_info())

        return cls(target=worker_thread, **kwds)

    def __init__(self, target: Target, *, queue_factory: Callable[[], Queue] = Queue):
        """Constructor.

        Parameters:
            target: the function to call in a separate thread. It will receive
                a Queue_ as its first argument and a function as its second
                argument. The thread should first call the function when it has
                performed any preparations that the caller should wait for. The
                thread should then regularly retrieve items from the queue and
                stop immediately if it receives `None`.
            queue_factory: a callable that constructs a new Queue_ instance when
                invoked with no arguments. You may use it to pass your own
                Queue_ subclass if needed.
        """
        self._queue = None
        self._task_group = None
        self._value = None

        self._queue_factory = queue_factory
        self._target = target

    def _notify_thread_started(self, value: Any = None) -> None:
        if self._value is None:
            raise RuntimeError("self._value must not be None here")
        else:
            from_thread.run_sync(self._value.set, value)

    async def __aenter__(self) -> Callable[[T], None]:
        if self._task_group is not None:
            raise RuntimeError("thread is already running")

        self._queue = self._queue_factory()
        self._value = AwaitableValue()

        self._task_group = create_task_group()
        await self._task_group.__aenter__()

        success = False
        result: Optional[T] = None
        try:
            self._task_group.start_soon(
                to_thread.run_sync,
                self._target,
                self._queue,
                self._notify_thread_started,
            )
            result = await self._value.wait()
            success = True
        finally:
            if not success:
                await self._task_group.__aexit__(*exc_info())
                self._task_group = None

        return result if result is not None else self._queue.put_nowait  # type: ignore

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        if self._queue:
            self._queue.put(None)
            self._queue = None

        try:
            if self._task_group:
                try:
                    return bool(
                        await self._task_group.__aexit__(exc_type, exc_value, tb)
                    )
                finally:
                    self._task_group = None
            else:
                return False
        finally:
            self._value = None



from contextlib import contextmanager
from typing import Callable
from time import time

__all__ = ("timing",)


class TimerContext:
    """Object returned from the `timing` context manager."""

    def __init__(self, timer: Callable[[], float]):
        """Constructor.

        Parameters:
            timer: a function that returns a (preferably monotonic) timestamp in
                seconds when called with no arguments
        """
        self._timer = timer
        self._start = timer()

    @property
    def elapsed(self) -> float:
        """Returns the number of seconds elapsed since entering the context."""
        return self._timer() - self._start


@contextmanager
def timing(description: str = "", timer: Callable[[], float] = time):
    """Context manager that allows us to measure the execution time of a
    code block.

    Parameters:
        description: textual description of what we are measuring
        timer: function that must be called with no arguments to get a
            reading of the clock we are using for measurements
    """
    context = TimerContext(timer)
    yield context
    if description:
        format_str = "{0}: {1:.3f}s"
        print(format_str.format(description, context.elapsed))



"""Classes related to accessing the localization subsystem of a Crazyflie."""

from enum import IntEnum
from struct import Struct
from typing import ClassVar, Iterable, Optional, Sequence, Tuple, Union

from aiocflib.crtp import CRTPPort
from aiocflib.utils.quaternion import compress_unit_quaternion, QuaternionXYZW

from .crazyflie import Crazyflie

__all__ = ("Localization",)


#: Maximum number of supported Lighthouse base stations
NUM_LIGHTHOUSE_BASE_STATIONS = 16


class LocalizationChannel(IntEnum):
    """Enum representing the names of the channels in the localization service
    of the CRTP protocol.
    """

    EXTERNAL_POSITION = 0
    GENERIC = 1
    POSITION_PACKED = 2


class GenericLocalizationCommand(IntEnum):
    """Enum representing the names of the commands in the generic channel of the
    localization service of the CRTP protocol.
    """

    RANGE_STREAM_REPORT = 0
    RANGE_STREAM_REPORT_FP16 = 1
    LPP_SHORT_PACKET = 2
    ENABLE_EMERGENCY_STOP = 3
    RESET_EMERGENCY_STOP_TIMEOUT = 4
    COMM_GNSS_NMEA = 6
    COMM_GNSS_PROPRIETARY = 7
    EXT_POSE = 8
    EXT_POSE_PACKED = 9
    LH_ANGLE_STREAM = 10
    LH_PERSIST_DATA = 11


class Localization:
    """Class representing the handler of messages related to the localization
    subsystem of a Crazyflie instance.
    """

    _crazyflie: Crazyflie

    _external_position_struct: ClassVar[Struct] = Struct("<fff")
    _external_position_packed_item_struct: ClassVar[Struct] = Struct("<Bhhh")
    _external_pose_struct: ClassVar[Struct] = Struct("<Bfffffff")
    _external_pose_packed_item_struct: ClassVar[Struct] = Struct("<BhhhL")
    _lpp_short_packet_struct: ClassVar[Struct] = Struct("<BB")
    _lighthouse_angle_struct: ClassVar[Struct] = Struct("<Bfhhhfhhh")
    _lighthouse_persist_struct: ClassVar[Struct] = Struct("<HH")

    @classmethod
    def encode_external_position_packed(
        cls, items: Sequence[Tuple[int, Tuple[float, float, float]]]
    ) -> bytes:
        """Encodes the payload of a "packed external position" packet that
        contains position information for multiple Crazyflie drones.

        The packet is intended to be broadcast to all Crazyflies in a network
        using the ``send_packet()`` method of a `Broadcaster` object.

        Parameters:
            items: a sequence of pairs containing a numeric Crazyflie ID
                (the last byte of its radio address) and a 3D coordinate.
                Coordinates must be less than ~32.7 meters in absolute value.
                At most four items fit into a single Crazyflie CRTP packet.
        """
        return b"".join(
            cls._external_position_packed_item_struct.pack(
                id, int(x * 1000), int(y * 1000), int(z * 1000)
            )
            for id, (x, y, z) in items
        )

    @classmethod
    def encode_external_pose_packed(
        cls, items: Sequence[Tuple[int, Tuple[float, float, float], QuaternionXYZW]]
    ) -> bytes:
        """Encodes the payload of a "packed external pose" packet that
        contains position and attitude information for multiple Crazyflie drones.

        The packet is intended to be broadcast to all Crazyflies in a network
        using the ``send_packet()`` method of a `Broadcaster` object.

        Parameters:
            items: a sequence of triplets containing a numeric Crazyflie ID
                (the last byte of its radio address), a 3D coordinate and a
                4D quaternion in XYZW order. Coordinates must be less than ~32.7
                meters in absolute value. At most two items fit into a single
                Crazyflie CRTP packet.
        """
        return b"".join(
            cls._external_pose_packed_item_struct.pack(
                id,
                int(x * 1000),
                int(y * 1000),
                int(z * 1000),
                compress_unit_quaternion(quat),
            )
            for id, (x, y, z), quat in items
        )

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the localization
                subsystem related messages
        """
        self._crazyflie = crazyflie

    async def _send_packet(
        self,
        data: Union[int, bytes],
        channel: LocalizationChannel = LocalizationChannel.GENERIC,
    ) -> None:
        await self._crazyflie.send_packet(
            port=CRTPPort.LOCALIZATION, channel=channel, data=data
        )

    async def send_external_position(
        self,
        x: Union[float, Sequence[float]],
        y: Optional[float] = None,
        z: Optional[float] = None,
    ) -> None:
        """Sends position information originating from an external positioning
        system into the Crazyflie.

        Parameters:
            x: the X coordinate. May also be a full 3D position vector; in this
                case y and z must be `None`.
            y: the Y coordinate
            z; the Z coordinate
        """
        if y is None and z is None:
            if isinstance(x, Iterable):
                data = self._external_position_struct.pack(*x)
            else:
                raise TypeError(
                    "x must be a sequence of floats when y and z are not given"
                )
        else:
            data = self._external_position_struct.pack(x, y, z)

        await self._send_packet(
            data,
            channel=LocalizationChannel.EXTERNAL_POSITION,
        )

    async def send_external_pose(
        self, pos: Sequence[float], quat: Sequence[float]
    ) -> None:
        """Sends pose (position and attitude) information originating from an
        external positioning system into the Crazyflie.

        Parameters:
            pos: the position vector (x, y, z)
            quat: the attitude quaternion (qx, qy, qz, qw)
        """
        x, y, z = pos
        qx, qy, qz, qw = quat
        await self._send_packet(
            self._external_pose_struct.pack(
                GenericLocalizationCommand.EXT_POSE, x, y, z, qx, qy, qz, qw
            ),
        )

    async def send_lpp_short_packet(self, dest_id: int, data: bytes) -> bool:
        """Sends an LPP short packet to the Loco Positioning System, using the
        Crazyflie as a proxy.

        Parameters:
            dest_id: ID of the Loco Positioning System node to send the packet to
            data: the raw LPP short packet to send

        Returns:
            whether the LPP short packet response indicated a success or a failure
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.LOCALIZATION,
            channel=LocalizationChannel.GENERIC,
            command=GenericLocalizationCommand.LPP_SHORT_PACKET,
            data=self._lpp_short_packet_struct.pack(dest_id) + data,
        )
        return len(response) > 0 and bool(response[0])

    async def enable_emergency_stop(self) -> None:
        """Sends an "enable emergency stop" packet to the Crazyflie."""
        await self._send_packet(
            GenericLocalizationCommand.ENABLE_EMERGENCY_STOP,
        )

    async def reset_emergency_stop_timeout(self) -> None:
        """Sends a "reset emergency stop timeout" packet to the Crazyflie to
        prevent it from stopping when the emergency stop watchdog is enabled.
        """
        await self._send_packet(
            GenericLocalizationCommand.RESET_EMERGENCY_STOP_TIMEOUT,
        )

    async def persist_lighthouse_data(
        self,
        geo_list: Optional[Iterable[int]] = None,
        calib_list: Optional[Iterable[int]] = None,
    ) -> bool:
        """Instructs the Crazyflie to persist the currently estimated geometry
        and calibration data of the Lighthouse subsystem into permanent storage.

        Parameters:
            geo_list: IDs of the Lighthouse base stations (0-based) whose
                geometry data must be persisted. Defaults to all stations when
                omitted.
            calib_list: IDs of the Lighthouse base stations (0-based) whose
                calibration data must be persisted. Defaults to the same value
                as the geometry list when omitted.

        Returns:
            whether the data was persisted successfully
        """
        if geo_list is None:
            geo_list = range(NUM_LIGHTHOUSE_BASE_STATIONS)

        if calib_list is None:
            calib_list = geo_list

        if not _is_valid_lighthouse_base_station_id_list(geo_list):
            raise ValueError("Geometry base station ID list is invalid")
        if not _is_valid_lighthouse_base_station_id_list(calib_list):
            raise ValueError("Calibration base station ID list is invalid")

        geo_mask, calib_mask = 0, 0
        for id in geo_list:
            geo_mask |= 1 << id
        for id in calib_list:
            calib_mask |= 1 << id

        response = await self._crazyflie.run_command(
            port=CRTPPort.LOCALIZATION,
            channel=LocalizationChannel.GENERIC,
            command=GenericLocalizationCommand.LH_PERSIST_DATA,
            data=self._lighthouse_persist_struct.pack(geo_mask, calib_mask),
        )

        return len(response) > 0 and bool(response[0])


def _is_valid_lighthouse_base_station_id_list(ids: Iterable[int]) -> bool:
    return all(
        isinstance(id, int) and id >= 0 and id < NUM_LIGHTHOUSE_BASE_STATIONS
        for id in ids
    )



"""Middleware that logs all incoming and outgoing packets of a CRTP driver
instance.
"""

from aiocflib.crtp.crtpstack import CRTPPacket, CRTPPort
from hexdump import dump, hexdump

from .base import MiddlewareBase
from .registry import register


try:
    from colorama import init, Fore, Back, Style

    _has_colors = True
except ImportError:

    class Unstyled:
        def __getattr__(self, attr):
            return ""

    Fore = Back = Style = Unstyled()

    def init():
        pass

    _has_colors = False


#: Underlined style extension for colorama
UNDERLINED = "\033[4m"


@register("log")
class LoggingMiddleware(MiddlewareBase):
    """Middleware that logs all incoming and outgoing packets of a CRTP driver
    instance.
    """

    _in = Fore.CYAN + Style.BRIGHT + "\u25c0" + Style.RESET_ALL
    _out = Fore.CYAN + "\u25b6" + Style.RESET_ALL
    _unsafe_in = Fore.RED + Style.BRIGHT + "\u25c0" + Style.RESET_ALL
    _unsafe_out = Fore.RED + "\u25b6" + Style.RESET_ALL

    @staticmethod
    def _format_type(packet: CRTPPacket) -> str:
        """Formats the port and channel of the given CRTP packet in the format
        that will be used in the log.
        """
        port, channel = packet.port, packet.channel
        try:
            code = port.code
        except Exception:
            code = str(port)
        return "{0:3}:{1}".format(code, channel)

    @staticmethod
    def _has_command_byte(packet: CRTPPacket) -> bool:
        """Returns whether the first byte of the data of the given CRTP packet
        is to be interpreted as a command byte (which is a common pattern in
        many CRTP services and channels).

        Parameters:
            packet: the CRTP packet to test

        Returns:
            whether the CRTP packet has a command byte in the first byte of the
            data payload
        """
        if len(packet.data) < 1:
            return False

        if packet.port == CRTPPort.LOGGING:
            return packet.channel in (0, 1)
        elif packet.port == CRTPPort.PARAMETERS:
            return packet.channel in (0, 3)
        elif packet.port == CRTPPort.MEMORY:
            return packet.channel == 0
        elif packet.port == CRTPPort.LOCALIZATION:
            return packet.channel == 1
        elif packet.port == CRTPPort.GENERIC_COMMANDER:
            return packet.channel == 0
        elif packet.port == CRTPPort.HIGH_LEVEL_COMMANDER:
            return packet.channel == 0
        elif packet.port == CRTPPort.PLATFORM:
            return packet.channel == 1
        elif packet.port == CRTPPort.DEBUG:
            return packet.channel == 0

        return False

    @staticmethod
    def _hexdump(data: bytes) -> str:
        """Creates a hex dump from the given data.

        Parameters:
            data: the data to format as hex dump

        Returns:
            the formatted hex dump, indented properly for the log
        """
        result = []
        for line in hexdump(data, result="generator"):
            _, _, line = line.partition(" ")
            result.append(line)
        return "\n             ".join(result)

    def _init(self) -> None:
        from aiocflib.crtp.drivers.cpplink import CppRadioDriver
        from aiocflib.crtp.drivers.radio import RadioDriver
        from aiocflib.crtp.drivers.sitl import SITLDriver
        from aiocflib.crtp.drivers.usb import USBDriver

        self._num_null_packets = 0
        driver = self._wrapped

        if isinstance(driver, USBDriver):
            index = driver.index
            abbreviation = "usb{0}".format(index if index is not None else "?")
        elif isinstance(driver, RadioDriver):
            address = driver.address
            abbreviation = dump(address[-2:], sep="") if address else "rdio"
        elif isinstance(driver, CppRadioDriver):
            address = driver.address
            abbreviation = dump(address[-2:], sep="") if address else "rdio"
        elif isinstance(driver, SITLDriver):
            abbreviation = "sitl"
        else:
            abbreviation = "????"

        abbreviation = Fore.MAGENTA + abbreviation + Style.RESET_ALL
        self._abbreviation = "{0:4}".format(abbreviation)

    def _report_null_packets(self) -> None:
        """Reports on the console that a certain number of null packets have
        been received.

        Parameters:
            bool: whether the link is currently safe
        """
        if self._num_null_packets == 1:
            print(
                "{0} {1} {2}Null packet{3}".format(
                    self._abbreviation,
                    self._in if self.is_safe else self._unsafe_in,
                    Style.DIM,
                    Style.RESET_ALL,
                )
            )
        elif self._num_null_packets > 1:
            print(
                "{0} {1} {2}{3} null packets{4}".format(
                    self._abbreviation,
                    self._in if self.is_safe else self._unsafe_in,
                    Style.DIM,
                    self._num_null_packets,
                    Style.RESET_ALL,
                )
            )
        self._num_null_packets = 0

    async def receive_packet(self) -> CRTPPacket:
        packet = await super().receive_packet()

        if packet.is_null:
            self._num_null_packets += 1
        else:
            if self._num_null_packets:
                self._report_null_packets()
            data = self._hexdump(packet.data)
            if self._has_command_byte(packet):
                data = Fore.YELLOW + UNDERLINED + data[:2] + Style.RESET_ALL + data[2:]
            type = Fore.GREEN + self._format_type(packet) + Style.RESET_ALL
            print(
                "{0} {1} {2} {3}".format(
                    self._abbreviation,
                    self._in if self.is_safe else self._unsafe_in,
                    type,
                    data,
                )
            )

        return packet

    async def send_packet(self, packet: CRTPPacket):
        if self._num_null_packets:
            self._report_null_packets()
        type = Fore.GREEN + self._format_type(packet) + Style.RESET_ALL
        data = self._hexdump(packet.data)
        if self._has_command_byte(packet):
            data = Fore.YELLOW + UNDERLINED + data[:2] + Style.RESET_ALL + data[2:]
        print(
            "{0} {1} {2} {3}".format(
                self._abbreviation,
                self._out if self.is_safe else self._unsafe_out,
                type,
                data,
            )
        )
        return await super().send_packet(packet)



"""Classes related to handling console messages of a Crazyflie."""

from typing import AsyncIterator, List

from anyio import fail_after

from aiocflib.crtp import CRTPPacket, CRTPPort
from aiocflib.utils.concurrency import aclosing

from .crazyflie import Crazyflie

__all__ = ("Console",)


class Console:
    """Class representing the handler of console messages for a Crazyflie
    instance.
    """

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the console
                messages
        """
        self._crazyflie = crazyflie

    async def messages(self, timeout: float = 1, partial_message_marker: str = "…"):
        """Async generator that yields full console messages from a
        Crazyflie.

        This generator essentially re-assembles the individual console message
        packets into full lines.

        Parameters:
            timeout: maximum number of seconds to wait for a newline character
                after a console message. If no newline character is received
                in this timeframe after a console message, the message will
                be posted separately, followd by a partial message marker
            partial_message_marker: marker to append to messages if a newline
                was not received in time after having received the message
        """
        partial_message_marker_bytes = partial_message_marker.encode("UTF-8")
        if not partial_message_marker_bytes.endswith(b"\n"):
            partial_message_marker_bytes += b"\n"

        parts: List[bytes] = []
        gen = self.packets()

        async with aclosing(gen):
            while True:
                packet = None

                try:
                    if not parts:
                        packet = await gen.__anext__()
                    else:
                        try:
                            with fail_after(timeout):
                                packet = await gen.__anext__()
                        except TimeoutError:
                            packet = None
                except StopAsyncIteration:
                    break

                if packet is not None:
                    data = packet.data.rstrip(b"\x00")
                else:
                    data = partial_message_marker_bytes

                while True:
                    data, sep, rest = data.partition(b"\n")
                    if data:
                        parts.append(data)
                    if sep:
                        yield (
                            b"".join(parts).decode("UTF-8", errors="backslashreplace")
                        )
                        parts.clear()
                        data = rest
                    else:
                        break

    async def packets(self) -> AsyncIterator[CRTPPacket]:
        """Async generator that yields console message packets from a Crazyflie,
        without reassembling them to full messages.
        """
        async for packet in self._crazyflie.packets(port=CRTPPort.CONSOLE):
            yield packet


async def test():
    uri = "radio+log://0/80/2M/E7E7E7E709"
    # uri = "sitl+log://"
    # uri = "usb+log://0"

    async with Crazyflie(uri) as cf:
        async for message in cf.console.messages():
            print(message)


if __name__ == "__main__":
    from aiocflib.crtp import init_drivers
    import trio

    init_drivers()
    try:
        trio.run(test)
    except KeyboardInterrupt:
        pass



"""Classes related to setting the colors and effects on the LED ring of a
Crazyflie.
"""

from anyio import sleep
from contextlib import asynccontextmanager
from enum import IntEnum
from typing import AsyncIterator, Optional, Union

from aiocflib.utils.colors import ColorLike, to_color

from .crazyflie import Crazyflie

__all__ = ("LEDRing", "LEDRingEffect")


class LEDRingEffect(IntEnum):
    BLACK = 0
    WHITE_SPIN = 1
    COLOR_SPIN = 2
    TILT = 3
    BRIGHTNESS = 4
    SPIN = 5
    DOUBLE_SPIN = 6
    SOLID_COLOR = 7
    LED_TEST = 8
    BATTERY_CHARGE = 9
    BOAT = 10
    SIREN = 11
    GRAVITY_LIGHT = 12
    VIRTUAL_MEM = 13
    FADE_COLOR = 14
    RSSI = 15
    LOCATION_SERVICE_STATUS = 16
    TIME_MEM = 17
    LIGHTHOUSE = 18


class LEDRing:
    """Class representing the LED ring of a Crazyflie instance."""

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle LED ring-related
                commands
        """
        self._crazyflie = crazyflie

    async def flash(self) -> None:
        """Instructs the LED ring to show a brief light signal to attract
        attention.
        """
        await self._crazyflie.parameters.set("system.highlight", 1)

    async def get_effect(self) -> Union[LEDRingEffect, int]:
        """Returns the current effect code of the LED ring.

        Returns:
            the current effect code as a member of the LEDRingEffect enum,
            if the effect code is known to the enum, otherwise as an integer
        """
        value = await self._crazyflie.parameters.get("ring.effect")
        value = int(value)
        try:
            return LEDRingEffect(value)
        except ValueError:
            return value

    async def is_installed(self) -> bool:
        """Returns whether the LED ring is installed on the Crazyflie."""
        await self._crazyflie.parameters.validate()
        return bool(await self._crazyflie.parameters.get("deck.bcLedRing"))

    async def set_color(self, color: ColorLike) -> None:
        """Sets a solid color on the LED ring.

        This command also switches the mode (effect) of the LED ring.
        """
        resolved_color = to_color(color)
        await self._crazyflie.parameters.set("ring.fadeTime", 0)
        await self._crazyflie.parameters.set("ring.fadeColor", resolved_color.rgb888)
        await self.set_effect(LEDRingEffect.FADE_COLOR)

    async def set_effect(self, effect: LEDRingEffect) -> None:
        """Switches the LED ring to the given light effect mode."""
        await self._crazyflie.parameters.set("ring.effect", effect)

    async def set_headlight(self, on: bool = True) -> None:
        """Turns the headlight on or off."""
        await self._crazyflie.parameters.set("ring.headlightEnable", 1 if on else 0)

    @asynccontextmanager
    async def set_effect_and_restore(
        self, effect: LEDRingEffect, old_effect: Optional[LEDRingEffect] = None
    ) -> AsyncIterator[None]:
        """Context manager that sets the LED ring effect to the given light
        effect when entering the context and restores it when exiting the context.
        """
        async with self._crazyflie.parameters.set_and_restore(
            "ring.effect", effect, old_effect
        ):
            yield

    async def test(self, duration: float = 2) -> None:
        """Tests the LED ring by sending it to testing mode for the given number
        of seconds.

        Parameters:
            duration: the duration of the test, in seconds
        """
        async with self.set_effect_and_restore(LEDRingEffect.LED_TEST):
            await sleep(duration)
        await self.set_headlight(False)

    async def turn_off(self) -> None:
        """Turns off the LED ring."""
        await self.set_effect(LEDRingEffect.BLACK)
        await self.set_headlight(False)

    @asynccontextmanager
    async def use(self) -> AsyncIterator[None]:
        """Context manager that turns off the LED ring when exiting the context."""
        try:
            yield
        finally:
            await self.turn_off()


async def test():
    from anyio import sleep

    uri = "radio+log://0/80/2M/E7E7E7E701"
    # uri = "sitl+log://"
    # uri = "usb+log://0"

    async with Crazyflie(uri, cache="/tmp/cfcache") as cf:
        installed = await cf.led_ring.is_installed()
        print(f"LED ring installed: {installed}")
        await sleep(1)

        async with cf.led_ring.use():
            print("Flashing LED ring...")
            await cf.led_ring.flash()
            await sleep(1)

            print("Testing LED ring...")
            await cf.led_ring.test()
            await sleep(1)

            print("Setting LED ring to solid colors...")
            await cf.led_ring.set_color("green")
            await sleep(1)
            await cf.led_ring.set_color((0, 127, 255))
            await sleep(1)
            await cf.led_ring.set_color("#ffcc00")
            await sleep(1)


if __name__ == "__main__":
    from aiocflib.crtp import init_drivers
    import trio

    init_drivers()
    try:
        trio.run(test)
    except KeyboardInterrupt:
        pass



"""Classes related to sending low-level roll-pitch-yaw-thrust and setpoint
messages to a Crazyflie.
"""

from enum import IntEnum
from struct import Struct
from typing import ClassVar

from aiocflib.crtp import CRTPPort

from .crazyflie import Crazyflie

__all__ = ("Commander",)


class SetpointType(IntEnum):
    """Enum representing the setpoint types that we can send to the generic
    commander port.
    """

    STOP = 0
    VELOCITY_WORLD = 1
    Z_DISTANCE = 2
    CPPM = 3
    ALTITUDE_HOLD = 4
    HOVER = 5
    FULL_STATE = 6
    POSITION = 7


class Commander:
    """Class responsible for sending low-level roll-pitch-yaw-thrust and
    setpoint messages to a Crazyflie instance.
    """

    _crazyflie: Crazyflie

    _altitude_hold_setpoint_struct: ClassVar[Struct] = Struct("<Bffff")
    _hover_setpoint_struct: ClassVar[Struct] = Struct("<Bffff")
    _position_setpoint_struct: ClassVar[Struct] = Struct("<Bffff")
    _rpyt_setpoint_struct: ClassVar[Struct] = Struct("<fffH")
    _velocity_world_setpoint_struct: ClassVar[Struct] = Struct("<Bffff")
    _z_distance_setpoint_struct: ClassVar[Struct] = Struct("<Bffff")

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie to which we need to send the messages
        """
        self._crazyflie = crazyflie

    async def send_altitude_hold_setpoint(
        self,
        roll: float = 0.0,
        pitch: float = 0.0,
        yaw_rate: float = 0.0,
        z_velocity: float = 0.0,
    ) -> None:
        """Sends an altitude hold setpoint to the Crazyflie; velocity is
        specified along the Z axis, while the XY axes are controlled by roll
        and pitch angles.

        Roll and pitch are in degrees. Yaw rate is in degrees/s. Z velocity is
        in m/s.
        """
        data = self._altitude_hold_setpoint_struct.pack(
            SetpointType.ALTITUDE_HOLD, roll, pitch, yaw_rate, z_velocity
        )
        await self._crazyflie.send_packet(port=CRTPPort.GENERIC_COMMANDER, data=data)

    async def send_hover_setpoint(
        self,
        vx: float = 0.0,
        vy: float = 0.0,
        yaw_rate: float = 0.0,
        z_distance: float = 0.0,
    ) -> None:
        """Sends a hover setpoint to the Crazyflie; velocity is specified in the
        XY plane, along with the desired distance from the ground and the yaw
        rate.

        Velocity components are in m/s. Yaw rate is in degrees/s. Z distance is
        in meters.
        """
        data = self._hover_setpoint_struct.pack(
            SetpointType.HOVER, vx, vy, yaw_rate, z_distance
        )
        await self._crazyflie.send_packet(port=CRTPPort.GENERIC_COMMANDER, data=data)

    async def send_position_setpoint(
        self,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 0.0,
        yaw: float = 0.0,
    ) -> None:
        """Sends a position setpoint to the Crazyflie with a fixed X-Y-Z
        coordinate triplet and a yaw angle.

        Coordinates are in meters; yaw is in degrees.
        """
        data = self._position_setpoint_struct.pack(SetpointType.POSITION, x, y, z, yaw)
        await self._crazyflie.send_packet(port=CRTPPort.GENERIC_COMMANDER, data=data)

    async def send_setpoint(
        self, roll: float, pitch: float, yaw: float, thrust: int
    ) -> None:
        """Sends a low-level roll-pitch-yaw-thrust setpoint to the
        Crazyflie.

        Thrust is automatically capped between 0 and 65535. Roll, pitch and
        yaw are in degrees.
        """
        if thrust > 0xFFFF:
            thrust = 0xFFFF
        elif thrust < 0:
            thrust = 0

        data = self._rpyt_setpoint_struct.pack(roll, -pitch, yaw, thrust)
        await self._crazyflie.send_packet(port=CRTPPort.COMMANDER, data=data)

    async def send_stop_setpoint(self) -> None:
        """Sends an immediate stop command, stopping the motors and potentially
        making the Crazyflie fall down and crash.
        """
        await self._crazyflie.send_packet(
            port=CRTPPort.GENERIC_COMMANDER, data=SetpointType.STOP
        )

    async def send_velocity_world_setpoint(
        self, vx: float = 0.0, vy: float = 0.0, vz: float = 0.0, yaw_rate: float = 0.0
    ) -> None:
        """Sends a low-level velocity setpoint in the world coordinate frame to
        the Crazyflie.

        Thrust is automatically capped between 0 and 65535. Velocity components
        are in m/s.
        """
        data = self._velocity_world_setpoint_struct.pack(
            SetpointType.VELOCITY_WORLD, vx, vy, vz, yaw_rate
        )
        await self._crazyflie.send_packet(port=CRTPPort.GENERIC_COMMANDER, data=data)

    async def send_z_distance_setpoint(
        self,
        roll: float = 0.0,
        pitch: float = 0.0,
        yaw_rate: float = 0.0,
        z_distance: float = 0.0,
    ) -> None:
        """Sends a low-level setpoint where the height is defined as an absolute
        setpoint along with the desired roll and pitch angles and the yaw rate.

        Roll and pitch are in degrees; yaw rate is in degrees/s.
        """
        data = self._z_distance_setpoint_struct.pack(
            SetpointType.Z_DISTANCE, roll, pitch, yaw_rate, z_distance
        )
        await self._crazyflie.send_packet(port=CRTPPort.GENERIC_COMMANDER, data=data)

    stop = send_stop_setpoint



"""Classes related to accessing the parameters subsystem of a Crazyflie."""

from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import dataclass
from enum import IntEnum
from errno import ENOENT
from struct import Struct, error as StructError
from typing import cast, Dict, List, Optional, Tuple, Union

from aiocflib.crtp import CRTPPort
from aiocflib.errors import error_to_string
from aiocflib.utils.toc import TOCCache, fetch_table_of_contents_gracefully

from .crazyflie import Crazyflie

__all__ = ("Parameters",)


class ParameterChannel(IntEnum):
    """Enum representing the names of the channels of the parameter service in
    the CRTP protocol.
    """

    TABLE_OF_CONTENTS = 0
    READ = 1
    WRITE = 2
    MISC = 3


class ParameterTOCCommand(IntEnum):
    """Enum representing the names of the table-of-contents commands in the
    parameter service of the CRTP protocol.

    These commands are valid for ParameterChannel.TABLE_OF_CONTENTS (i.e.
    channel 0).
    """

    RESET = 0
    READ_PARAMETER_DETAILS_V2 = 2
    READ_TOC_INFO_V2 = 3


class ParameterCommand(IntEnum):
    """Enum representing the names of the generic commands in the parameter
    service of the CRTP protocol.

    These commands are valid for ParameterChannel.MISC (i.e. channel 3).
    """

    SET_BY_NAME = 0
    VALUE_UPDATED = 1
    GET_EXTENDED_TYPE = 2
    PERSISTENT_STORE = 3
    PERSISTENT_GET_STATE = 4
    PERSISTENT_CLEAR = 5
    GET_DEFAULT_VALUE = 6


#: Dictionary mapping integer type codes to their C types, Python structs and
#: aliases
_type_properties = {
    # C type, Python struct, aliases
    0x00: ("int8_t", Struct("<b"), ("int8", "i8")),
    0x01: ("int16_t", Struct("<h"), ("int16", "i16")),
    0x02: ("int32_t", Struct("<i"), ("int32", "i32")),
    0x03: ("int64_t", Struct("<q"), ("int64", "i64")),
    0x05: ("fp16", Struct("<h"), ()),
    0x06: ("float", Struct("<f"), ()),
    0x07: ("double", Struct("<d"), ()),
    0x08: ("uint8_t", Struct("<B"), ("uint8", "u8")),
    0x09: ("uint16_t", Struct("<H"), ("uint16", "u16")),
    0x0A: ("uint32_t", Struct("<L"), ("uint32", "u32")),
    0x0B: ("uint64_t", Struct("<Q"), ("uint64", "u64")),
}


class ParameterType(IntEnum):
    """Enum containing the possible types of a parameter and the corresponding
    numeric identifiers.
    """

    INT8 = 0
    INT16 = 1
    INT32 = 2
    INT64 = 3
    FLOAT = 6
    DOUBLE = 7
    UINT8 = 8
    UINT16 = 9
    UINT32 = 10
    UINT64 = 11

    @classmethod
    def to_type(cls, value: Union[int, str, "ParameterType"]):
        """Converts an integer, string or ParameterType_ instance to a
        ParameterType.
        """
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            return _type_names[value]
        else:
            return ParameterType(value)

    @property
    def aliases(self) -> Tuple[str]:
        """Returns the registered type aliases of this type."""
        return (_type_properties[self][0],) + _type_properties[self][2]

    @property
    def length(self) -> int:
        """Returns the number of bytes that a single value of this log
        variable would occupy.
        """
        return self.struct.size

    @property
    def struct(self) -> Struct:
        """Returns a Python struct that can be used to encode or decode
        parameter values of this type.
        """
        return _type_properties[self][1]

    def encode_value(self, value) -> bytes:
        """Encodes a single value of this parameter type into its raw byte-level
        representation.
        """
        return self.struct.pack(value)


#: Type specification for objects that can be converted into a parameter type
ParameterTypeLike = Union[str, int, ParameterType]

#: Dictionary mapping string type aliases to types
_type_names = {alias: type for type in ParameterType for alias in type.aliases}


@dataclass(frozen=True)
class ParameterSpecification:
    """Class representing the specification of a single parameter of the
    Crazyflie.
    """

    id: int
    """The numeric identifier of the parameter."""

    type: int
    """The type of the parameter; typically one of the values from ParameterType_"""

    group: str
    """Name of the group that the parameter is a part of."""

    name: str
    """Name of the parameter within its group."""

    read_only: bool
    """Whether the parameter is read-only."""

    has_extended_info: bool
    """Whether the parameter has corresponding extended information."""

    @classmethod
    def from_bytes(cls, data: bytes, id: int):
        try:
            type = data[0] & 0x0F
            read_only = bool(data[0] & 0x40)
            has_extended_info = bool(data[0] & 0x10)
            group, name, *rest = data[1:].split(b"\x00")
            return cls(
                id=id,
                type=type,
                group=group.decode("ascii"),
                name=name.decode("ascii"),
                read_only=read_only,
                has_extended_info=has_extended_info,
            )
        except Exception:
            raise ValueError("invalid parameter description") from None

    def encode_value(self, value) -> bytes:
        """Encodes a single value of this parameter into its raw byte-level
        representation.
        """
        return _type_properties[self.type][1].pack(value)

    @property
    def encoded_length(self) -> int:
        """Returns the number of bytes that will be used to encode a parameter
        of this type.
        """
        return _type_properties[self.type][1].size

    @property
    def full_name(self) -> str:
        """Returns the fully-qualified name of the parameter, which is
        the concatenation of the group and the name of the parameter, separated
        by a dot.
        """
        return f"{self.group}.{self.name}"

    def parse_value(self, data: bytes) -> Union[int, float]:
        """Parses the raw byte-level representation of a single value of this
        parameter, as received from the Crazyflie, and returns the corresponding
        Python value.
        """
        return _type_properties[self.type][1].unpack(data)[0]

    def to_bytes(self) -> bytes:
        header = (
            (int(self.type) & 0x0F)
            + (0x40 if self.read_only else 0)
            + (0x10 if self.has_extended_info else 0)
        )

        parts = []
        parts.append(bytes((header,)))
        parts.append(self.group.encode("ascii"))
        parts.append(b"\x00")
        parts.append(self.name.encode("ascii"))
        parts.append(b"\x00")

        return b"".join(parts)


@dataclass
class PersistentParamState:
    """Class representing the state of a persisted parameter on the Crazyflie."""

    default_value: float
    """The default value of the parameter in the Crazyflie firmware."""

    stored_value: Optional[float]
    """The value of the parameter stored in the permanent storage of the
    Crazyflie; `None` if the parameter is not persisted.
    """

    @property
    def is_stored(self) -> bool:
        """Whether the permanent storage on the Crazyflie has a value associated
        to this parameter.
        """
        return self.stored_value is not None


class Parameters:
    """Class representing the handler of messages related to the parameter
    subsystem of a Crazyflie instance.
    """

    _cache: Optional[TOCCache]
    _crazyflie: Crazyflie

    _values: Dict[str, Union[int, float]]
    _variables: List[ParameterSpecification]
    _variables_by_name: Dict[str, ParameterSpecification]

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the parameter
                subsystem related messages
        """
        self._cache = crazyflie._get_cache_for("param_toc")
        self._crazyflie = crazyflie

        self._values = None  # type: ignore
        self._variables = None  # type: ignore
        self._variables_by_name = None  # type: ignore

    async def clear_persisted_value(self, name: str) -> None:
        """Clears the persisted value of a parameter, given its fully-qualified
        name.

        Parameters:
            name: the fully-qualified name of the parameter
        """
        await self.validate()

        parameter = self._variables_by_name[name]
        index = parameter.id

        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.MISC,
            command=(ParameterCommand.PERSISTENT_CLEAR, index & 0xFF, index >> 8),
        )

        if len(response) < 1:
            raise ValueError("invalid response for clearing a persisted parameter")

        if response[0]:
            raise RuntimeError(
                f"failed to clear persisted value of parameter {name}; code = {response[0]}"
            )

    async def get(self, name: str, fetch: bool = False) -> Union[int, float]:
        """Returns the current value of a parameter, given its fully-qualified
        name.

        Parameters:
            name: the fully-qualified name of the parameter
            fetch: whether to forcefully fetch a new value from the drone even
                if we have a locally cached copy

        Returns:
            the current value of the parameter (which may be a cached value)
        """
        value = (
            None
            if fetch
            else self._values.get(name)
            if self._values is not None
            else None
        )
        if value is None:
            value = self._values[name] = await self._fetch(name)
        return value

    async def get_default(self, name: str) -> Union[int, float]:
        """Returns the default value of a parameter, given its fully-qualified
        name.

        Parameters:
            name: the fully-qualified name of the parameter

        Returns:
            the default value of the parameter
        """
        await self.validate()

        parameter = self._variables_by_name[name]
        if parameter.read_only:
            raise RuntimeError("read-only parameters have no default value")

        index = parameter.id
        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.MISC,
            command=(ParameterCommand.GET_DEFAULT_VALUE, index & 0xFF, index >> 8),
        )

        if not response:
            raise IndexError("parameter index out of range")
        if len(response) < 1:
            raise ValueError("invalid response for parameter query")
        if response[0]:
            if response[0] == ENOENT:
                raise RuntimeError(
                    f"parameter {name} does not exist or has no default value"
                )
            else:
                raise RuntimeError(
                    f"error while retrieving default value of parameter {name}"
                )
        if len(response) < 2:
            raise ValueError("invalid response for parameter query")
        return parameter.parse_value(response[1:])

    async def get_persistence_state(self, name: str) -> PersistentParamState:
        """Returns the persistence state of a parameter, given its
        fully-qualified name.

        Parameters:
            name: the fully-qualified name of the parameter

        Returns:
            an object storing whether the parameter is persisted, and if so,
            what is its default and persisted value
        """
        await self.validate()

        parameter = self._variables_by_name[name]
        index = parameter.id
        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.MISC,
            command=(ParameterCommand.PERSISTENT_GET_STATE, index & 0xFF, index >> 8),
        )

        if not response:
            raise IndexError("parameter index out of range")
        if len(response) < 1:
            raise ValueError("invalid response for parameter persistence state query")
        if len(response) == 1:
            if response[0] == ENOENT:
                raise RuntimeError(f"parameter {name} does not exist")
            else:
                raise RuntimeError(
                    f"error {response[0]} while retrieving persistence state of parameter {name}"
                )

        is_persisted = bool(response[0])
        num_bytes = parameter.encoded_length + 1
        if is_persisted:
            num_bytes += parameter.encoded_length

        if len(response) < num_bytes:
            raise ValueError("invalid response for parameter query (too short)")

        default_value = parameter.parse_value(
            response[1 : (parameter.encoded_length + 1)]
        )
        if is_persisted:
            stored_value = parameter.parse_value(
                response[(parameter.encoded_length + 1) :]
            )
        else:
            stored_value = None

        return PersistentParamState(
            default_value=default_value, stored_value=stored_value
        )

    async def has(self, name: str) -> bool:
        """Returns whether the parameter with the given name is known to the
        Crazyflie.

        Parameters:
            name: the fully-qualified name of the parameter

        Returns:
            True if the parameter is known to the Crazyflie, False otherwise
        """
        await self.validate()
        return self.has_sync(name)
        return name in self._variables_by_name

    def has_sync(self, name: str) -> bool:
        """Synchronous variant of the ``has()`` method; returns whether the
        parameter with the given name is known to the Crazyflie.

        This method assumes that the parameter TOC has already been downloaded
        with ``self.validate()``.

        Parameters:
            name: the fully-qualified name of the parameter

        Returns:
            True if the parameter is known to the Crazyflie, False otherwise
        """
        return name in self._variables_by_name

    async def persist(self, name: str) -> None:
        """Stores the current value of the given parameter in persistent
        storage, given its fully-qualified name.

        Parameters:
            name: the fully-qualified name of the parameter
        """
        await self.validate()

        parameter = self._variables_by_name[name]
        index = parameter.id

        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.MISC,
            command=(ParameterCommand.PERSISTENT_STORE, index & 0xFF, index >> 8),
        )

        if len(response) < 1:
            raise ValueError("invalid response for persisting the value of a parameter")

        if response[0]:
            raise RuntimeError(f"failed to persist value of parameter {name}")

    async def set(self, name: str, value) -> None:
        """Sets the value of a parameter, given its fully-qualified name.

        Parameters:
            name: the fully-qualified name of the parameter
            value: the new value of the parameter
        """
        # TODO(ntamas): make it possible to set by name if we have no TOC
        await self.validate()

        parameter = self._variables_by_name[name]
        if parameter.read_only:
            raise AttributeError("{} is read only".format(name))

        index = parameter.id

        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.WRITE,
            command=(index & 0xFF, index >> 8),
            data=parameter.encode_value(value),
        )

        if not response:
            raise IndexError("parameter index out of range")
        if len(response) < 1:
            raise ValueError("invalid response for parameter setting")

        self._values[name] = parameter.parse_value(response)

    @asynccontextmanager
    async def set_and_restore(
        self, name: str, value, restore_to=None, fetch: bool = False
    ):
        """Asynchronous context manager that sets the value of a parameter
        when entering the context and restores it to another value when exiting
        the context.

        Parameters:
            name: the fully-qualified name of the parameter
            value: the new value of the parameter
            restore_to: the old value of the parameter to restore when exiting
                the context. `None` means to use the last cached copy of the
                parameter, or to fetch a new parameter if `fetch` is set to
                `True`.
            fetch: whether to fetch the old parameter value unconditionally before
                entering the context. Used only if `restore_to` is `None`,
                otherwise the value of `restore_to` will be unsed, in which
                case there is no point in fetching the old value.
        """
        if restore_to is None:
            restore_to = await self.get(name, fetch=fetch)

        try:
            await self.set(name, value)
            yield
        finally:
            await self.set(name, restore_to)

    async def set_fast(self, name: str, type: ParameterTypeLike, value) -> None:
        """Sets the value of a parameter without fetching the full parameter
        TOC first, given its fully-qualified name and its type.

        This function is useful if you don't want to spend time with fetching
        the parameter TOC from the drone and you only need to set the values
        of some parameters. Triggering a parameter read anywhere will download
        the TOC from the drone anyway if needed.
        """
        resolved_type = ParameterType.to_type(type)
        group, _, name_bytes = name.encode("ascii").rpartition(b".")
        command = [ParameterCommand.SET_BY_NAME, group, 0, name_bytes, 0]
        data = bytes((resolved_type,)) + resolved_type.encode_value(value)
        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.MISC,
            command=command,
            data=data,
        )

        if not response:
            raise ValueError("Crazyflie returned an empty response")

        code = response[0]
        if code:
            raise ValueError(
                "Crazyflie returned error code {0} ({1})".format(
                    code, error_to_string(code)
                )
            )

    async def trigger(self, name: str) -> None:
        """Triggers some function on the drone by setting the value of the
        corresponding parameter to 1.

        Some parameters on the Crazyflie are used to trigger some functionality;
        for instance, the ``kalman.resetEstimation`` parameter is used to reset
        the state of the Kalman filter. This function is a convenience wrapper
        for setting the value of this parameter to 1.
        """
        return await self.set(name, 1)

    async def validate(self) -> None:
        """Ensures that the basic information about the parameters of the
        Crazyflie are downloaded.
        """
        if self._variables is not None:
            return

        self._variables, self._variables_by_name = await self._validate()
        self._values = {}

    async def _fetch(self, name: str) -> Union[int, float]:
        """Furcefully fetch the current value of a parameter, given its
        fully-qualified name.

        Parameters:
            name: the fully-qualified name of the parameter

        Returns:
            the current value of the parameter
        """
        await self.validate()

        parameter = self._variables_by_name[name]
        index = parameter.id

        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.READ,
            command=(index & 0xFF, index >> 8),
        )

        if not response:
            raise IndexError("parameter index out of range")
        if len(response) < 2:
            raise ValueError("invalid response for parameter query")

        return parameter.parse_value(response[1:])

    async def _get_parameter_spec_by_index(self, index: int) -> ParameterSpecification:
        """Returns the specification of the parameter with the given index.

        Parameters:
            index: the index of the parameter to retrieve

        Returns:
            the specification of the parameter with the given index
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.TABLE_OF_CONTENTS,
            command=(
                ParameterTOCCommand.READ_PARAMETER_DETAILS_V2,
                index & 0xFF,
                index >> 8,
            ),
        )
        if not response:
            raise IndexError("parameter index out of range")
        return ParameterSpecification.from_bytes(response, id=index)

    async def _get_table_of_contents_info(self) -> Tuple[int, int]:
        """Returns basic information about the table of contents of the
        parameter list, including the number of parameters and the CRC32
        hash of the parameter table.

        Returns:
            the number of parameters and the CRC32 hash of the parameter table
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.PARAMETERS,
            channel=ParameterChannel.TABLE_OF_CONTENTS,
            command=ParameterTOCCommand.READ_TOC_INFO_V2,
        )
        try:
            return cast(Tuple[int, int], Struct("<HI").unpack(response))
        except StructError:
            raise ValueError("invalid parameter TOC info response") from None

    async def _validate(self):
        """Downloads the basic information about the parameters of the
        Crazyflie.
        """
        # TODO(ntamas): when connecting to multiple drones with the same TOC,
        # fetch the parameters only from one of them
        parameters = await fetch_table_of_contents_gracefully(
            self._cache,
            self._get_table_of_contents_info,
            self._get_parameter_spec_by_index,
            ParameterSpecification.from_bytes,
            ParameterSpecification.to_bytes,
        )
        by_name = {parameter.full_name: parameter for parameter in parameters}
        return parameters, by_name



from typing import Generator, Tuple


def chunkify(
    addr: int, length: int, step: int
) -> Generator[Tuple[int, int], None, None]:
    """Calculates the start addresses and the sizes of individual chunks
    when trying to read some data from (or write some data to) a given start
    address with the given total length, and each individual request may
    have a limited size only.

    Parameters:
        addr: the address to start reading from or writing to
        length: the total number of bytes to read or write
        step: the number of bytes that we can read or write in a single
            request

    Returns:
        a generator yielding address-length combinations for the individual
        read requests that we need to execute
    """
    end = addr + length
    for start in range(addr, end, step):
        yield start, min(step, end - start)



"""Statistics-related utility classes and functions."""

from collections import deque
from typing import Optional

__all__ = ("SlidingWindowMean",)


class SlidingWindowMean:
    """Sliding window mean calculator."""

    def __init__(self, window_size: int, fill: Optional[float] = None):
        """Constructor."""
        self._fill = fill
        self._window_size = window_size
        self.reset()

    def add(self, value: float) -> float:
        if len(self._data) == self._window_size:
            self._sum += value - self._data[0]
        else:
            self._sum += value
        self._data.append(value)
        return self._sum

    @property
    def mean(self) -> float:
        return self._sum / len(self._data)

    def reset(self) -> None:
        """Resets the sliding window mean calculator to its base state."""
        n = self._window_size
        if self._fill is not None:
            self._data = deque([self._fill] * n, n)
        else:
            self._data = deque([], n)
        self._sum = float(sum(self._data))



from .crtpstack import (
    CRTPCommandLike,
    CRTPDataLike,
    CRTPDispatcher,
    CRTPPacket,
    CRTPPort,
    CRTPPortLike,
    LinkControlChannel,
    MemoryType,
)
from .device import CRTPDevice
from .drivers import CRTPDriver, init_drivers

__all__ = (
    "CRTPCommandLike",
    "CRTPDataLike",
    "CRTPDevice",
    "CRTPDispatcher",
    "CRTPDriver",
    "CRTPPacket",
    "CRTPPort",
    "CRTPPortLike",
    "LinkControlChannel",
    "MemoryType",
    "init_drivers",
)
